""" this module prepares the local file paths for data and results"""

from typing import Dict
import logging
import sys
from pathlib import Path

# from .. import __package_name__

logger = logging.getLogger(__name__)

from raman_fitting.config import filepath_settings as config

#%%


def get_directory_paths_for_run_mode(run_mode: str = "", **kwargs) -> Dict:
    """
    Parameters
    ----------
    run_mode : str, optional
        this is name of the run mode. The default is ''.
    **kwargs : TYPE
        kwargs can contain keys such as DATASET_DIR to overwrite the standard config paths.

    Returns
    -------
    dest_dirs : dict
        dict containing 3 keys [RESULTS_DIR, DATASET_DIR, INDEX_FILE]

    """
    dest_dirs = {}
    DATASET_DIR = None
    RESULTS_DIR = None

    if run_mode in ("DEBUG", "testing"):
        # self.debug = True
        RESULTS_DIR = config.TESTS_RESULTS_DIR
        DATASET_DIR = config.TESTS_DATASET_DIR

    elif run_mode == "make_examples":
        RESULTS_DIR = config.PACKAGE_HOME.joinpath("example_results")
        DATASET_DIR = config.TESTS_DATASET_DIR
        # self._kwargs.update({'default_selection' : 'all'})

    elif run_mode in ("normal", "make_index"):
        RESULTS_DIR = config.RESULTS_DIR
        DATASET_DIR = config.DATASET_DIR
        # INDEX_FILE = config.INDEX_FILE
    else:
        logger.warning(f"Run mode {run_mode} not recognized. Exiting...")

    INDEX_FILE = RESULTS_DIR / config.INDEX_FILE_NAME
    DB_FILE = RESULTS_DIR / config.DB_FILE_NAME

    dest_dirs = {
        "RESULTS_DIR": Path(RESULTS_DIR),
        "DATASET_DIR": Path(DATASET_DIR),
        "INDEX_FILE": Path(INDEX_FILE),
        "DB_FILE": Path(DB_FILE),
    }

    if kwargs:
        dest_dirs = override_from_kwargs(dest_dirs, **kwargs)

    check_and_make_dirs(dest_dirs)

    return dest_dirs


def check_and_make_dirs(dest_dirs: dict = {}):
    DATASET_DIR = dest_dirs.get("DATASET_DIR", None)
    if DATASET_DIR:
        create_dataset_dir(DATASET_DIR)
    else:
        logger.warning(f"No datafiles directory was set for . Exiting...")

    RESULTS_DIR = dest_dirs.get("RESULTS_DIR", None)
    if RESULTS_DIR:
        if not RESULTS_DIR.is_dir():
            RESULTS_DIR.mkdir(exist_ok=True, parents=True)
            logger.info(
                f"check_and_make_dirs the results directory did not exist and was created at:\n{RESULTS_DIR}\n"
            )
    """ returns index file path """


def override_from_kwargs(_dict, **kwargs):
    _kwargs = kwargs
    if _kwargs:
        _keys = [i for i in _dict.keys() if i in _kwargs.keys()]
        _new_dict = {
            k: Path(val) if not _kwargs.get(k, None) else _kwargs[k]
            for k, val in _dict.items()
        }
        if _new_dict != _dict:
            logger.debug(f"override_from_kwargs keys {_keys} were overwritten")
        return _new_dict
    else:
        return _dict


def create_dataset_dir(DATASET_DIR):  # pragma: no cover
    if not DATASET_DIR.is_dir():
        logger.warning(
            f"The datafiles directory does not exist yet, the program will now try to create this folder.\n{DATASET_DIR}"
            # therefore {config.__package_name__} can not find any files.
            # The program will now try to create this folder.
        )
        try:
            DATASET_DIR.mkdir()
            logger.warning(
                f"""The datafiles directory has now been created at:
{DATASET_DIR}
please place your raman datafiles in this folder and run {config.__package_name__} again.
{config.__package_name__} exits now.
"""
            )
            sys.exit()
            # IDEA build in daemon version with folder watcher....
        except Exception as exc:
            logger.warning(
                f"""The datafiles directory could not be created at:
{DATASET_DIR}
An unexpected error ocurred:
{exc}
please redefine the path for dataset_dir in the config settings.
"""
            )
    else:
        # Check if dir is not empty else raise a warning
        _diter = DATASET_DIR.iterdir()
        try:
            next(_diter)
        except StopIteration:
            logger.warning(
                f"""The datafiles directory is empty:
{DATASET_DIR}
please place your files in here or
change this path in the config settings.
"""
            )



"""Collection of method for parsing a filename"""
# -*- coding: utf-8 -*-

import datetime
from typing import Tuple

# class ParserMethods

__all__ = ["filestem_to_sid_and_pos", "sID_to_sgrpID", "get_fstats"]


def filestem_to_sid_and_pos(stem: str, seps=("_", " ", "-")) -> Tuple[str, str]:
    """
    Parser for the filenames -> finds SampleID and sample position

    Parameters
    ----------
    # ramanfile_stem : str
    #    The filepath which the is parsed
    seps : tuple of str default
        ordered collection of seperators tried for split
        default : ('_', ' ', '-')

    Returns
    -------
    tuple of strings
        Collection of strings which contains the parsed elements.
    """

    split = None
    first_sep_match_index = min(
        [n for n, i in enumerate(seps) if i in stem], default=None
    )
    first_sep_match = (
        seps[first_sep_match_index] if first_sep_match_index is not None else None
    )
    split = stem.split(first_sep_match)
    _lensplit = len(split)

    if _lensplit == 0:
        sID, position = split[0], 0
    elif len(split) == 1:
        sID, position = split[0], 0
    elif len(split) == 2:
        sID = split[0]
        _pos_strnum = "".join(i for i in split[1] if i.isnumeric())
        if _pos_strnum:
            position = int(_pos_strnum)
        else:
            position = split[1]
    elif len(split) >= 3:
        sID = "_".join(split[0:-1])
        position = int("".join(filter(str.isdigit, split[-1])))
    #                split =[split_Nr0] + [position]
    return (sID, position)
    # else:
    #     sID = split[0] # default take split[0]
    #     if ''.join(((filter(str.isdigit,split[-1])))) == '':
    #         position = 0
    #     else:
    #         position = int(''.join(filter(str.isdigit,split[-1])))


def sID_to_sgrpID(sID: str, max_len=4) -> str:
    """adding the extra sample Group key from sample ID"""

    _len = len(sID)
    _maxalphakey = min(
        [n for n, i in enumerate(sID) if not str(i).isalpha()], default=_len
    )
    _maxkey = min((_len, _maxalphakey, max_len))
    sgrpID = "".join([i for i in sID[0:_maxkey] if i.isalpha()])
    return sgrpID


def get_fstats(fstat) -> Tuple:
    """converting creation time and last mod time to datetime object"""
    c_t = fstat.st_ctime
    m_t = fstat.st_mtime
    c_tdate, m_tdate = c_t, m_t

    try:
        c_t = datetime.datetime.fromtimestamp(fstat.st_ctime)
        m_t = datetime.datetime.fromtimestamp(fstat.st_mtime)
        c_tdate = c_t.date()
        m_tdate = m_t.date()
    except OverflowError as e:
        pass
    except OSError as e:
        pass

    return c_tdate, c_t, m_tdate, m_t, fstat.st_size

    # def _extra_sID_check_if_reference(self, ref_ID = 'Si-ref'):
    #     if ref_ID in self.stem:
    #         position = 0
    #         sID = 'Si-ref'
    #         return (sID, position)
    #     else:
    #         return None



"""
Created on Sun May 30 12:35:58 2021

@author: DW
"""
if __name__ == "__main__":
    from model_validation import PeakModelValidator

    from raman_fitting.config.config import PACKAGE_HOME

else:
    from model_validation import PeakModelValidator

    from ..config.config import PACKAGE_HOME


class ModelConfigurator:

    standard_config_file = "model_config_standard.cfg"

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def find_user_config_files(self):
        pass

    def file_handler(self):
        pass

    def standard_valid_models(self):

        peak_collection = PeakModelValidator()



""" Default peaks used for 1st order deconvolution"""

if __name__ == "__main__":
    from raman_fitting.deconvolution_models.default_peaks.base_peak import BasePeak
else:
    from .base_peak import BasePeak

__all__ = ["G_peak", "D_peak", "D2_peak", "D3_peak", "D5_peak", "Si1_peak"]

# ====== FIRST ORDER PEAKS ======= #


class G_peak(metaclass=BasePeak):

    """
    Graphite belongs to the P63/mmc (D46h) space group. If considering only a graphene plane, at
    the Ã point of the Brillouin zone, there are six normal modes that possess only one mode (doubly
    degenerate in plane) with a E2g representation, which is Raman active
    G ; Ideal graphitic lattice (E2g-symmetry)
    G peak center stable over different laser wavelengths.
    Influenced by potential, HSO4 adsorption (or ionization of G- and G+),
    magnetic fields, pressure
    Für G: 1580-1590 D5 und D2 weiß ich nicht
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "G"
        self.input_param_settings = {
            "center": {"value": 1571, "min": 1545, "max": 1595},
            "sigma": {"value": 30, "min": 5, "max": 150},
            "amplitude": {"value": 35, "min": 5, "max": 500},
        }


class D_peak(metaclass=BasePeak):
    """
    D or D1 ; Disordered graphitic lattice (graphene layer edges,A1gsymmetry)
    A defective graphite presents other bands that can be as intense as the G band at D=1350 and D'=1615 cm-1
    These bands are activated by defects due to the breaking of the crystal symmetry that relax the Raman selection rules.
    Für D1: 1340-1350
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D"
        self.input_param_settings = {
            "center": {"value": 1350, "min": 1330, "max": 1380},
            "sigma": {"value": 35, "min": 1, "max": 150},
            "amplitude": {"value": 120, "min": 1e-05, "max": 500},
        }


class D2_peak(metaclass=BasePeak):
    """
    D2 or D' ; Right next to the G peak, sometimes not obvious as G peak split.
    Disordered graphitic lattice (surface graphene layers,E2g-symmetry)
    j.molstruc.2010.12.065
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D2"
        self.input_param_settings = {
            "center": {"value": 1606, "min": 1592, "max": 1635},
            "sigma": {"value": 30, "min": 5, "max": 150},
            "amplitude": {"value": 35, "min": 5, "max": 500},
        }


class D3_peak(metaclass=BasePeak):
    """
    D3 or D'' or A or Am ; Between the D and G peak, sometimes too broad.
    For amorphous carbon (Gaussian[26]or Lorentzian[3,18,27]line shape).
    Für D3: 1495-1515
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D3"
        self.input_param_settings = {
            "center": {"value": 1480, "min": 1450, "max": 1525},
            "sigma": {"value": 25, "min": 1, "max": 150},
            "amplitude": {"value": 25, "min": 1e-02, "max": 500},
        }


class D4_peak(metaclass=BasePeak):
    """
    D4 or I ; Below D band, a shoulder sometimes split with D5 band.
    Disordered graphitic lattice (A1gsymmetry)[10],polyenes[3,27], ionic impurities
    D4 peak at 1212 cm−1
    Jurkiewicz, K., Pawlyta, M., Zygadło, D. et al. J Mater Sci (2018) 53: 3509. https://doi.org/10.1007/s10853-017-1753-7
    Für D4: 1185-1210, but depends on if there is D5 or not.
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D4"
        self.input_param_settings = {
            "center": {"value": 1230, "min": 1180, "max": 1310},
            "sigma": {"value": 40, "min": 1, "max": 150},
            "amplitude": {"value": 20, "min": 1e-02, "max": 200},
        }


class D5_peak(metaclass=BasePeak):
    """
    D5 peak at 1110 cm−1. At lowest should of D peak, below D4.
    Ref: Jurkiewicz, K., Pawlyta, M., Zygadło, D. et al. J Mater Sci (2018) 53: 3509. https://doi.org/10.1007/s10853-017-1753-7
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D5"
        self.input_param_settings = {
            "center": {"value": 1150, "min": 1100, "max": 1200},
            "sigma": {"value": 40, "min": 1, "max": 250},
            "amplitude": {"value": 20, "min": 1e-02, "max": 200},
        }


class Si1_peak(metaclass=BasePeak):
    """
    ===== Extra peak at ca. 960 cm-1 presumably from Si substrate 2nd order === not from Nafion...
    => Either cut the Spectra 1000-2000
    => Place an extra Gaussian peak at 960 in the fit
    """

    def __init__(self):
        self.peak_type = "Gaussian"
        self.peak_name = "Si1"
        self.input_param_settings = {
            "center": {"value": 960, "min": 900, "max": 980},
            "sigma": {"value": 10, "min": 0, "max": 150},
            "amplitude": {"value": 10, "min": 0, "max": 200},
        }


def test_for_Si_substrate(
    model,
):  # IDEA test fit on spectrum to decide wether Si substrate is required or not
    """make test fit for only slice 900-1000"""



if __name__ == "__main__":
    from raman_fitting.deconvolution_models.base_peak import BasePeak
else:
    from .base_peak import BasePeak

__all__ = ["D4D4_peak", "D1D1_peak", "GD1_peak", "D2D2_peak"]

# ====== SECOND ORDER PEAKS ======= #
class D4D4_peak(metaclass=BasePeak):
    """
    2nd order D4 peak
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D4D4"
        self.input_param_settings = {
            "center": {"value": 2435, "min": 2400, "max": 2550},
            "sigma": {"value": 30, "min": 1, "max": 200},
            "amplitude": {"value": 2, "min": 1e-03, "max": 100},
        }


class D1D1_peak(metaclass=BasePeak):
    """
    2nd order D(1) peak, aka 2D
    2450 cm􀀀1 band, which has been attributed recently to a D + D”
    band by Couzi et al. [61], the D + D’ (in literature, the wrong D + G label is often found [62]), the
    2D’ bands and the 2D + G band.
    1627 and 2742 cm􀀀1 bands as the D’ and 2D
    """

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D1D1"
        self.input_param_settings = {
            "center": {"value": 2650, "min": 2600, "max": 2750},
            "sigma": {"value": 60, "min": 1, "max": 200},
            "amplitude": {"value": 14, "min": 1e-03, "max": 100},
        }


class GD1_peak(metaclass=BasePeak):
    """
    2nd order G+D(1) peak
    """

    #        D2D2_mod = VoigtModel(prefix='D2D2_')
    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "GD1"
        self.input_param_settings = {
            "center": {"value": 2900, "min": 2800, "max": 2950},
            "sigma": {"value": 50, "min": 1, "max": 200},
            "amplitude": {"value": 10, "min": 1e-03, "max": 100},
        }


class D2D2_peak(metaclass=BasePeak):
    """
    2nd order D2 peak, aka 2D2
    """

    #        D2D2_mod = VoigtModel(prefix='D2D2_')

    def __init__(self):
        self.peak_type = "Lorentzian"
        self.peak_name = "D2D2"
        self.input_param_settings = {
            "center": {"value": 3250, "min": 3000, "max": 3400},
            "sigma": {"value": 60, "min": 20, "max": 200},
            "amplitude": {"value": 1, "min": 1e-03, "max": 100},
        }


# ====== SECOND ORDER PEAKS ======= #



# from .logging_config import get_console_handler
import logging
import pathlib
from sys import exit

from raman_fitting import __package_name__

logger = logging.getLogger(__name__)


# import pandas as pd
# pd.options.display.max_rows = 10
# pd.options.display.max_columns = 10
# __package_name__ = 'raman_fitting'

CONFIG_FILE = pathlib.Path(__file__).resolve()
PACKAGE_ROOT = CONFIG_FILE.parent.parent
MODEL_DIR = PACKAGE_ROOT / "deconvolution_models"

# TESTS_ROOT_DIR = PACKAGE_ROOT.parent.parent / "tests"
# TESTS_ROOT_DIR =
TESTS_DATASET_DIR = PACKAGE_ROOT / "datafiles" / "example_files"

# Home dir from pathlib.Path for storing the results
PACKAGE_HOME = (
    pathlib.Path.home() / f".{__package_name__}"
)  # pyramdeconv is the new version package name


try:
    if not PACKAGE_HOME.is_dir():
        try:
            logger.warning(
                f"Package home directory did not exist, will now be created at:\n{PACKAGE_HOME}\n--------------------"
            )
            PACKAGE_HOME.mkdir()
        except Exception as exc:
            logger.warning(
                f"Package home mkdir unexpected error\n{exc}.\nFolder{PACKAGE_HOME} could not be created, exiting."
            )
            exit()
    else:
        logger.info(
            f"Package home directory exists at:\n{PACKAGE_HOME}\n--------------------"
        )
except Exception as exc:
    logger.warning(
        f"Unexpected error with checking for package home folder:\nFolder:{PACKAGE_HOME}\nError:\n{exc}\n {__package_name__} can not run."
    )
    exit()


TESTS_RESULTS_DIR = PACKAGE_HOME / "test_results"

DATASET_DIR = PACKAGE_HOME / "datafiles"
RESULTS_DIR = PACKAGE_HOME / "results"

# Optional local configuration file
LOCAL_CONFIG_FILE = PACKAGE_HOME / "local_config.py"

# ADAPT to your own configurations
if 0:  # pragma: no cover
    # IDEA implement config file handling
    # LOCAL_CONFIG_FILE.is_file():
    try:
        # PACKAGE_ROOT, MODEL_DIR are not locally configurated
        from .local_config import DATASET_DIR, RESULTS_DIR

        logger.info(
            f" Importing settings from local config...",
            "\n",
            f"RESULTS_DIR : {RESULTS_DIR}",
            "\n",
            f"From file: {LOCAL_CONFIG_FILE}",
        )

    except Exception as e:
        logger.warning(
            f"Failed importing settings from local config...{LOCAL_CONFIG_FILE} because {e}"
        )

# Storage file of the index
INDEX_FILE_NAME = f"{__package_name__}_index.csv"
DB_FILE_NAME = f"{__package_name__}_sqlite.db"
# INDEX_FILE = RESULTS_DIR / INDEX_FILE_NAME

# import configparser
# config = configparser.ConfigParser()
# config['DEFAULT'] = {'A': 1}



"""
Overview of commands used for testing and coverage
run commands examples
        pytest --cov=app
    --cov check if it is folder or package
    pytest --cov=fizzbuzz
    --cov-report=term-missing
        shows the terms that are missing

    --cov-branch branch
        coverage, where conditionals
        require a double  coverage

    use # pragma: no cover
        to iqnore coverage test for that code block

    pytest --cov --cov-report=term-missing

with coverage only:
    coverage run -m unittest discover
    coverage combine
    coverage xml
    coverage report -m
    coverage report -m --skip-covered

"""



# flake8: noqa
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Configuration file for pytest and commonly used fixtures
"""
import importlib

import pandas
import pytest

# rname = importlib.util.resolve_name('raman_fitting', None)
# importlib.import_module('raman_fitting')
# print(f"pytest: {__name__},file: {__file__}\n name:")
# Incremental tests


def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed (%s)" % previousfailed.name)


# Global fixtures



import logging
import sys

# Multiple calls to logging.getLogger('someLogger') return a
# reference to the same logger object.  This is true not only
# within the same module, but also across modules as long as
# it is in the same Python interpreter process.

FORMATTER = logging.Formatter(
    "%(asctime)s — %(name)s — %(levelname)s —" "%(funcName)s:%(lineno)d — %(message)s"
)



log_format = (
    "[%(asctime)s] — %(name)s — %(levelname)s —" "%(funcName)s:%(lineno)d—12s %(message)s")
    # '[%(asctime)s] %(levelname)-8s %(name)-12s %(message)s')

# Define basic configuration
logging.basicConfig(
    # Define logging level
    level=logging.DEBUG,
    # Define the format of log messages
    format=log_format,
    # Provide the filename to store the log messages
    filename=('debug.log'),
)


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 28 15:08:26 2021

@author: zmg
"""

import inspect
import logging
from collections import namedtuple
from itertools import groupby
from pathlib import Path
from typing import Tuple
from warnings import warn

import matplotlib.pyplot as plt
import pandas as pd
from lmfit import Parameters

# _file_parent_name = Path(__file__).parent.name
# print(__name__,__file__,f'name: {_file_parent_name}')

if __name__ == "__main__":  # or _file_parent_name == 'deconvolution_models':
    # import first_order_peaks
    # import second_order_peaks
    # import normalization_peaks
    from default_peaks import BasePeak
else:
    from .. import __package_name__
    from .default_peaks.base_peak import BasePeak

    logger = logging.getLogger(__package_name__)


#%%
class PeakValidationWarning(UserWarning):
    pass


class NotFoundAnyModelsWarning(PeakValidationWarning):
    pass


class CanNotInitializeModelWarning(PeakValidationWarning):
    pass


class PeakModelValidator:
    """
    This class collects all BasePeak (=BASE_PEAK) type classes, which are costum lmfit type models, and
    constructs an iterable collection of all defined Child class.
    Each subclass of BasePeak is:
        - validated: instance check
        - filtered: optional
        - sorted: sorting for valid models based on defined center position of BasePeak

    Followed by color assignment to each BasePeak and collection of lmfit_models

    """

    # _standard_modules = [first_order_peaks, second_order_peaks, normalization_peaks]
    BASE_PEAK = BasePeak

    ModelValidation = namedtuple(
        "ModelValidation", "valid peak_group model_inst message"
    )

    CMAP_OPTIONS_DEFAULT = ("Dark2", "tab20")
    fallback_color = (0.4, 0.4, 0.4, 1.0)

    debug = False

    # standard_model_selection = []
    # _bad_models = []
    # _skip_models = []
    # endwsith = '_peak'

    def __init__(self, *args, cmap_options=CMAP_OPTIONS_DEFAULT, **kwargs):
        # self.model_prefixes = model_prefixes
        # self._endswith = endwsith
        self.debug = self._set_debug(**kwargs)
        self._cmap_options = cmap_options

        self._inspect_models = self.get_subclasses_from_base(self.BASE_PEAK)

        self.valid_models = []
        self._invalid_models = []
        self.valid_models, self._invalid_models = self.validation_inspect_models(
            inspect_models=self._inspect_models
        )
        self.selected_models = self.filter_valid_models(self.valid_models)
        self.selected_models = self.sort_selected_models(self.selected_models)

        self.lmfit_models = self.assign_colors_to_lmfit_mod_inst(self.selected_models)
        self.add_model_names_var_names(self.lmfit_models)

        self.model_dict = self.get_model_dict(self.lmfit_models)
        self.options = self.model_dict.keys()

        # self.options = ()
        # self.extra_assignments()

    def _set_debug(self, **value):
        _debug = self.debug
        if isinstance(value, dict):
            if "debug" in value.keys():
                _debug = bool(value.get("debug", False))
        return _debug

    def get_subclasses_from_base(self, _BaseClass):
        """Finds subclasses of the BasePeak metaclass, these should give already valid models"""

        _all_subclasses = []
        if inspect.isclass(_BaseClass):
            if hasattr(_BaseClass, "subclasses"):
                _all_subclasses = _BaseClass.subclasses
            elif hasattr(_BaseClass, "__subclassess__"):
                _all_subclasses = _BaseClass.__subclasses__
            else:
                warn(
                    f"\nNo baseclasses were found for {str(_BaseClass)}:\n missing attributes",
                    NotFoundAnyModelsWarning,
                )
        else:
            warn(
                f"\nNo baseclasses were found for {str(_BaseClass)}:\n is not a class",
                NotFoundAnyModelsWarning,
            )

        if not _all_subclasses:
            warn(
                f"\nNo baseclasses were found in inspected modules for {str(_BaseClass)}:\n",
                NotFoundAnyModelsWarning,
            )

        return _all_subclasses

        # {", ".join(self._standard_modules)}
        # elif not self._inspect_models:
        # warn(f'\nNo base models were found in:\n {", ".join([str(i) for i in self._inspect_modules_all])}.\n', NotFoundAnyModelsWarning)
        # assert self._inspect_models, 'inspect.getmembers found 0 models, change the search parameters for _standard_modules or BASE_PEAK'

    def _inspect_modules_for_classes(self):
        """Optional method Inspect other modules for subclasses"""
        pass
        # self._inspect_modules_all = _all_subclasses
        # [cl for i in (inspect.getmembers(mod, inspect.isclass)
        # for mod in self._standard_modules)
        # for cl in i]

    def validation_inspect_models(self, inspect_models: list = []):
        """Validates each member of a list for making a valid model instance"""
        _model_validations = []
        valid_models = []

        for model in inspect_models:
            _module = model.__module__
            try:
                _succes, _inst, _msg = self.validate_model_instance(model)
            except Exception as e:
                _succes, _inst = False, model
                _msg = f"Unexpected error for validate model instance : {e}\n"
            finally:
                _args = (_succes, _module, _inst, _msg)
                _model_validations.append(self.ModelValidation(*_args))
                if self.debug:
                    print(_args)

        _invalid_models = [i for i in _model_validations if not i.valid]
        valid_models = [i for i in _model_validations if i.valid]

        if not valid_models:
            warn(
                f'\nNo valid models were found in:\n {", ".join([str(i) for i in inspect_models])}\
                \t\nOnly invalid models: {", ".join([str(i) for i in _invalid_models])}.\n',
                NotFoundAnyModelsWarning,
            )

        return valid_models, _invalid_models

    def filter_valid_models(self, value):
        """Optional method for extra filters on valid model selection"""
        return value
        # self._skipped_models = set(self._bad_models + self._skip_models)
        # if self.standard_model_selection:
        # self.selected_models = [i for i in self.valid_models if not i.model_inst.name in self._skipped_models]
        # if self._endswith:
        # self.selected_models = [(m, ngr) for m, ngr in self.valid_models
        # if (i.model_inst.name.endswith(self._endswith) and not i.model_inst.name  in self._skipped_models)]

    def sort_selected_models(self, value):
        """Sorting the selected valid models for color assigment etc.."""
        _sorted = value
        _setting_key = None
        try:
            _setting_key = [i for i in self.BASE_PEAK._fields if "param_hints" in i]
            if value:
                if _setting_key:
                    _sorted = sorted(
                        value,
                        key=lambda x: getattr(x.model_inst, _setting_key[0]).get(
                            "center", 0
                        ),
                    )
        except Exception as e:
            raise (f"Unable to sort:\n {value}\n{e}")
        finally:
            _sorted = sorted(_sorted, key=lambda x: x.peak_group)
            return _sorted

    def validate_model_instance(self, value):
        """
        Returns a boolean, model and message depending on the validation of the model class.
        Invalid classes can raise warnings, but exception only when no valid models are found.
        """

        try:
            if self.debug:
                print(f"validate model inst value:", value)
            _inst = value()
            if self.debug:
                print(f"validate model inst:", _inst)
        except Exception as e:
            _err = f"Unable to initialize model {value},\n{e}"
            warn(f"{_err}", CanNotInitializeModelWarning)
            return (False, value, _err)

        for field in self.BASE_PEAK._fields:
            if not hasattr(_inst, field):
                return (False, value, f"instance {_inst} has no attr {field}.\n")
            if not getattr(_inst, field):
                return (False, value, f"instance {_inst}, {field} is None.\n")
            if "param_hints" in field:
                _settings = getattr(_inst, field)
                _center = _settings.get("center", None)
                if not _center:
                    return (
                        False,
                        value,
                        f"instance {_inst}, settings {_settings} center is None.\n",
                    )
        return (True, _inst, f"{_inst} is a valid model")

    @staticmethod
    def get_cmap_list(
        lst, cmap_options: Tuple = (), fallback_color: Tuple = ()
    ) -> Tuple:

        cmap = [(0, 0, 0, 1) for i in lst]  # black as fallback default color

        # set fallback color from class
        if isinstance(fallback_color, tuple):
            if len(fallback_color) == 4:
                cmap = [fallback_color for i in lst]

        # set cmap colors from cmap options
        if cmap_options:

            try:
                pltcmaps = [plt.get_cmap(cmap) for cmap in cmap_options]
                # Take shortest colormap but not
                cmap = min(
                    [i for i in pltcmaps if len(lst) <= len(i.colors)],
                    key=lambda x: len(x.colors),
                    default=cmap,
                )
                # if succesfull
                if "ListedColormap" in str(type(cmap)):
                    cmap = cmap.colors

            except Exception as exc:
                logger.warning(f"get_cmap_list error setting cmap colors:{exc}")

        return cmap

    def assign_colors_to_lmfit_mod_inst(self, selected_models: list):
        cmap_get = self.get_cmap_list(
            selected_models,
            cmap_options=self._cmap_options,
            fallback_color=self.fallback_color,
        )
        lmfit_models = []
        for n, _arg in enumerate(selected_models):
            _m_inst = _arg.model_inst
            _m_inst._modelvalidation = _arg
            _m_inst.color = ", ".join([str(i) for i in cmap_get[n]])
            # _m_inst._funcname = str(m).split('__main__.')[-1][:-2]
            _m_inst._lenpars = len(_m_inst.peak_model.param_names)
            lmfit_models.append(_m_inst)
        return lmfit_models
        # self.lmfit_models= sorted(self.lmfit_models, key= lambda x: x._lenpars)
        # self.lmfit_models = _mod_inst

    def add_standard_init_params(self):
        self.standard_init_params = Parameters()
        self.standard_init_params.add_many(*BasePeak._params_guesses_base)

    def add_model_names_var_names(self, lmfit_models):
        _mod_param_names = {
            i.peak_model.name: i.peak_model.param_names for i in lmfit_models
        }
        return _mod_param_names

    def get_df_models_parameters(self):
        _models = pd.DataFrame(
            [
                (
                    i.model.name,
                    len(i.peak_model.param_names),
                    ", ".join(i.peak_model.param_names),
                )
                for i in self.lmfit_models
            ],
            columns=["Model_EEC", "model_lenpars", "model_parnames"],
        )
        return _models

    def get_model_dict(self, lmfit_models):
        model_dict = {i.__class__.__name__: i for i in lmfit_models}
        return model_dict

    def get_dict(self):
        return {
            i.__module__ + ", " + i.__class__.__name__: i for i in self.lmfit_models
        }

    def __getattr__(self, name):
        # raise AttributeError(f'Chosen name "{name}" not in in options: "{", ".join(self.options)}".')
        try:
            _options = self.__getattribute__("options")
            if name in _options:
                return self.model_dict.get(name, None)
            raise AttributeError(
                f'Chosen name "{name}" not in options: "{", ".join(_options)}".'
            )
        except AttributeError:
            # if 'normalization' in name:
            # return self.normalization_model()
            raise AttributeError(f'Chosen name "{name}" not in attributes')
        # else:
        # raise AttributeError(f'Chosen name "{name}" not in in options: "{", ".join(self.options)}".')

    def normalization_model(self):
        pass  # IDEA separate peaks in groups

    def __iter__(self):
        for mod_inst in self.lmfit_models:
            yield mod_inst
        # self.params_set = set([a for m in self.lmfit_models for a in m[1].param_names])

    #    lmfit_models = [Model(i.func,name=i.name) for i in model_selection]
    def __repr__(self):
        _repr = "Validated Peak model collection"
        if self.selected_models:
            _selmods = f", {len(self.selected_models)} models from: " + "\n\t- "
            _repr += _selmods
            _joinmods = "\n\t- ".join(
                [f"{i.peak_group}: {i.model_inst} \t" for i in self.selected_models]
            )
            _repr += _joinmods
        else:
            _repr += ", empty selected models"
        return _repr


if __name__ == "__main__":
    a = PeakModelValidator(debug=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 10 16:15:25 2021

@author: zmg
"""

import sqlite3

from raman_fitting.config import filepath_settings


class RamanDB:
    def __init__(self):
        self.dbpath = filepath_settings.RESULTS_DIR.joinpath("sqlite.db")

    def conn(self):
        self.conn = sqlite3.connect(self.dbpath)



# flake8: noqa
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 14 09:12:56 2021

@author: zmg
"""

import unittest

import pandas as pd
import pytest
from lmfit import Model

import raman_fitting
from raman_fitting.deconvolution_models.fit_models import Fitter, PrepareParams

# try:
#     import raman_fitting


# except Exception as e:
#     print(f'pytest file {__file__}, {__name__} error {e}')


class TestFitter(unittest.TestCase):
    def test_empty_Fitter(self):
        ft = Fitter({})
        self.assertFalse(ft.start_fit)
        self.assertEqual(ft.spectra, {})
        ft.fit_delegator()


class TestPrepareParams(unittest.TestCase):
    def test_empty_PrepareParams(self):
        # pp =
        with self.assertRaises(AttributeError):
            PrepareParams({})


def _testing():

    self = ft
    self = prep


if __name__ == "__main__":
    unittest.main()



# pylint: disable=W0614,W0611,W0622
# flake8: noqa
# isort:skip_file

__author__ = "David Wallace"
__docformat__ = "restructuredtext"
__status__ = "Development"
__future_package_name__ = "pyramdeconv"
__current_package_name__ = "raman_fitting"
__package_name__ = __current_package_name__


try:
    from ._version import version

    __version__ = version
except:
    __version__ = "__version__ = '0.6.15'"


from raman_fitting.config import filepath_settings
from raman_fitting.config import logging_config

# VERSION_PATH = config.PACKAGE_ROOT / 'VERSION.txt'
# with open(VERSION_PATH, 'r') as version_file:
# IDEA change version definitino
# __version__ = version_file.read().strip()

import logging
import sys
import warnings

# Configure logger for use in package
logger = logging.getLogger(__package_name__)

log_format = (
    "[%(asctime)s] — %(name)s — %(levelname)s —" "%(funcName)s:%(lineno)d—12s %(message)s")
    # '[%(asctime)s] %(levelname)-8s %(name)-12s %(message)s')

# Define basic configuration
logging.basicConfig(
    # Define logging level
    level=logging.DEBUG,
    # Define the format of log messages
    format=log_format,
    # Provide the filename to store the log messages
    filename=('debug.log'),
)

formatter = logging.Formatter(log_format)
# logger.setLevel(logging.DEBUG)
# from raman_fitting.config import logging_config
# logger.addHandler(logging_config.get_console_handler())
# logger.propagate = False

# create console handler
ch = logging.StreamHandler(stream=sys.stdout)
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)

# add the handlers to the logger
logger.addHandler(ch)

# This code is written for Python 3.
if sys.version_info[0] != 3:
    logger.error("raman_fitting requires Python 3.")
    sys.exit(1)

# Let users know if they're missing any hard dependencies
hard_dependencies = ("numpy", "pandas", "scipy", "matplotlib", "lmfit")
soft_dependencies = {}
missing_dependencies = []

import importlib

for dependency in hard_dependencies:
    if not importlib.util.find_spec(dependency):
        missing_dependencies.append(dependency)

if missing_dependencies:
    raise ImportError(f"Missing required dependencies {missing_dependencies}")

for dependency in soft_dependencies:
    if not importlib.util.find_spec(dependency):
        warnings.warn(
            f"Missing important package {dependency}. {soft_dependencies[dependency]}"
        )

del hard_dependencies, soft_dependencies, dependency, missing_dependencies

# from raman_fitting import main_run_fit, indexer

# Other user-facing functions
from .api import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May  3 11:10:59 2021

@author: zmg
"""

import copy
import logging
from collections import namedtuple
from dataclasses import dataclass, field
from operator import itemgetter
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd

# from .. import __package_name__

from raman_fitting.indexing.filedata_parser import SpectrumReader
from raman_fitting.processing.cleaner import (
    BaselineSubtractorNormalizer,
    Despiker,
    SpectrumMethods,
)
from raman_fitting.processing.spectrum_template import SpecTemplate, SpectrumWindows

# from .parser import Parser
logger = logging.getLogger(__name__)


@dataclass(order=True, frozen=False)
class SpectrumDataLoader:
    """
    Raman Spectrum Loader Dataclass, reads in the file and constructs a clean spectrum from the data.
    A sequence of steps is performed on the raw data from SpectrumReader.
    The steps/methods are: smoothening filter, despiking and baseline correction.
    """

    # IDEA Fix this class, simplify

    _fields = ("ramanshift", "intensity")
    # _spectrum_grp_cols = ['PAR_file','Segment #',EvRHE, 'RPM_DAC']
    file: Path = field(default=Path(Path.cwd().joinpath("empty.txt")))
    # sample_position: int = 0
    # ramanshift: np.array = field(default=np.array([]), init=False)
    # intensity: np.array = field(default=np.array([]), init=False)
    spectrum_length: int = field(default=0, init=False)
    info: Dict = field(default_factory=dict, repr=False)
    ovv: pd.DataFrame = field(default=pd.DataFrame(), repr=False)
    run_kwargs: Dict = field(default_factory=dict, repr=False)

    def __post_init__(self):

        self._qcnm = self.__class__.__qualname__

        self.register = {}  # this stores the data of each method as they are performed
        self.filtered_intensity = None
        self._despike = None
        self._baseline_corrected = None
        self.clean_data = None
        self.clean_df = {}  # dict of DataFrames
        self.register_df = pd.DataFrame()
        self._read_succes = False

        self.load_data_delegator()

    def register_spectrum(self, ramanshift, intensity, label):
        _spec = SpecTemplate(ramanshift, copy.deepcopy(intensity), label)
        self.register.update({label: _spec})

    def __getattr__(self, attr):
        """checks if attr is in instance dicts before raising error"""
        if attr in self.run_kwargs.keys():
            # FIXME CARE getting attributes from kwargs
            return self.run_kwargs.get(attr, None)
        elif attr in self.info.keys():
            return self.info.get(attr, None)
        else:
            raise AttributeError(
                f'Attribute "{attr}" is not in {self.run_kwargs.keys()} or {self.info.keys()}.'
            )

    def load_data_delegator(self):
        """calls the SpectrumReader class"""

        if self.info:
            FP_from_info = self.info.get("FilePath", None)
            if FP_from_info:
                if Path(FP_from_info) != self.file:
                    raise ValueError(
                        f"Mismatch in value for FilePath:\{self.file} != {FP_from_info}"
                    )
        else:
            self.info = {"FilePath": self.file}

        raw_spectrum = SpectrumReader(self.file)
        # print("======== ", raw_spectrum)

        self.register_spectrum(raw_spectrum.ramanshift, raw_spectrum.intensity, "raw")
        if raw_spectrum.spectrum_length > 0:
            self.spectrum_length = raw_spectrum.spectrum_length
            self.spectrum_methods_delegator()
        else:
            logger.warning(f"{self._qcnm} load data fail for:\n\t {self.file}")
        self.info = {**self.info, **self.run_kwargs}

    def spectrum_methods_delegator(self):
        self.filter_data(on_lbl="raw", out_lbl="filtered")
        self.despike(on_lbl="filtered", out_lbl="despiked")
        self.baseline_correction(on_lbl="despiked", out_lbl="clean_data")
        self.set_clean_data_df()
        self.set_df_from_register()

    def filter_data(self, on_lbl="raw", out_lbl="filtered"):
        _r, _int, _lbl = self.register.get(on_lbl)
        logger.debug(f"{self.file} try to filter len int({len(_int)}),({type(_int)})")
        filtered_intensity = SpectrumMethods.filtered_int(intensity=_int)
        self.filtered_intensity = filtered_intensity
        self.register_spectrum(_r, filtered_intensity, out_lbl)

    def despike(self, on_lbl="filtered", out_lbl="despiked"):
        _r, _int, _lbl = self.register.get(on_lbl)
        _despike = Despiker(_int)  # IDEA check for nan in array
        self._despike = _despike
        self.register_spectrum(_r, _despike.despiked_intensity, out_lbl)

    def baseline_correction(self, on_lbl="despiked", out_lbl="clean_data"):
        _r, _int, _lbl = self.register.get(on_lbl)
        _baseline_corrected = BaselineSubtractorNormalizer(_r, _int, label="despiked")
        self._baseline_corrected = _baseline_corrected

        _fullspec = _baseline_corrected.norm_data["full"]
        self.register_spectrum(_fullspec.ramanshift, _fullspec.intensity, out_lbl)
        self.clean_data = _baseline_corrected.norm_data

    def set_clean_data_df(self):

        self.clean_df = {
            k: pd.DataFrame(
                {"ramanshift": val.ramanshift, f"int_{self.SamplePos}": val.intensity}
            )
            for k, val in self.clean_data.items()
        }

    def set_df_from_register(self):
        _regdf = pd.DataFrame()
        for k, val in self.register.items():
            _spec = pd.DataFrame(
                {
                    "ramanshift": val.ramanshift,
                    f"{k}_int_{self.SampleID}_{self.SamplePos}": val.intensity,
                }
            )
            if _regdf.empty:
                _regdf = _spec
            else:
                _regdf = pd.merge_asof(_regdf, _spec, on="ramanshift")
        self.register_df = _regdf
        logger.debug(
            f"{self._qcnm} set_df_from_register len int({len(_regdf)}),({type(_regdf)})"
        )

    def plot_raw(self):
        _raw_lbls = [
            i
            for i in self.register_df.columns
            if not any(a in i for a in ["ramanshift", "clean_data"])
        ]
        self.register_df.plot(x="ramanshift", y=_raw_lbls)

    def split_data(self, on_lbl="filtered"):
        _r, _int, _lbl = self.register.get(on_lbl)  # unpacking data from register
        for windowname, limits in SpectrumWindows().items():
            ind = (_r >= np.min(limits)) & (_r <= np.max(limits))
            _intslice = _int[ind]
            label = f"{_lbl}_window_{windowname}"
            self.register_spectrum(_r, _intslice, label)


class SpectrumDataCollection:
    """
    This class takes in a collection of SpectrumDataLoader instances.
    It checks each member of the list and this enables the option
    to take the mean of several spectra from the same SampleID.
    """

    MeanSpecTemplate = namedtuple(
        "MeanSpectras", "windowname sID_rawcols sIDmean_col mean_info mean_spec"
    )

    def __init__(self, spectra: List = [SpectrumDataLoader]):
        self._qcnm = self.__class__.__qualname__
        self._spectra = spectra
        Validators.check_members(
            self._spectra
        )  # only raises warning when errors are found
        self.spectra = Validators.check_spectra_lengths(self._spectra)

        self.info = self.get_mean_spectra_info(self.spectra)
        self.info_df = pd.DataFrame(self.info, index=[0])
        self.prep_clean_data = self.get_mean_spectra_prep_data(self.spectra)

        self.calc_mean()

    @staticmethod
    def get_mean_spectra_info(spectra: List[SpectrumDataLoader]) -> Dict:
        """retrieves the info dict from spec instances and merges dict in keys that have 1 common value"""

        try:
            _all_spec_info = [spec.info for spec in spectra if hasattr(spec, "info")]

            _all_spec_info_merged = {
                k: val for i in _all_spec_info for k, val in i.items()
            }

            _all_spec_info_sets = [
                (k, set([i.get(k, None) for i in _all_spec_info]))
                for k in _all_spec_info_merged
            ]

            mean_spec_info = {
                k: list(val)[0] for k, val in _all_spec_info_sets if len(val) == 1
            }
        except Exception as exc:
            logger.warning(f"get_mean_spectra_info failed for spectra {spectra}")
            mean_spec_info = {}

        mean_spec_info.update({"mean_spectrum": True})

        return mean_spec_info

    @staticmethod
    def get_mean_spectra_prep_data(spectra: List[SpectrumDataLoader]) -> Dict:
        """retrieves the clean data from spec instances and makes lists of tuples"""
        # and merges dict in keys that have 1 common value'''
        try:
            _all_spec = [
                spec
                for spec in spectra
                if hasattr(spec, "clean_data") and hasattr(spec, "SamplePos")
            ]

            _all_spec_clean_data_keys = {
                k for i in _all_spec for k in i.clean_data.keys()
            }

            clean_prep_data = {
                k: [(i.SamplePos, i.clean_data.get(k, None)) for i in _all_spec]
                for k in _all_spec_clean_data_keys
            }
        except Exception as exc:
            logger.warning(f"get_mean_spectra_prep_data failed for spectra {spectra}")
            clean_prep_data = {}

        return clean_prep_data

    def calc_mean(self):
        """Core function of the merging of spectra of different sample positions"""
        _merged_window_specs = {}
        _speclst = []
        _posmean_lbl_base = f'int_{self.info.get("SampleID")}_mean'
        for wndwnm, data in self.prep_clean_data.items():

            _merge_df = pd.DataFrame()
            _pos_lbl_lst = []

            for _pos, _sp in data:
                _pos_lbl = f"int_{_pos}"

                _dfspec = pd.DataFrame(
                    {"ramanshift": _sp.ramanshift, _pos_lbl: _sp.intensity}
                )

                if _merge_df.empty:
                    _merge_df = _dfspec
                else:
                    _merge_df = pd.merge_asof(_merge_df, _dfspec, on="ramanshift")
                _pos_lbl_lst.append(_pos_lbl)

            _posmean_lbl = f"{_posmean_lbl_base}_{len(_pos_lbl_lst)}"
            _merge_df = _merge_df.assign(
                **{_posmean_lbl: _merge_df[_pos_lbl_lst].mean(axis=1)}
            )
            _merged_window_specs.update({wndwnm: _merge_df})

            _old_spec = self.MeanSpecTemplate(
                wndwnm, _pos_lbl_lst, _posmean_lbl, self.info_df, _merge_df
            )
            _speclst.append(_old_spec)

        self.fitting_spectra = _speclst
        self.mean_data = _merged_window_specs

    def __repr__(self):
        return f"{self.info}"


class Validators:
    """collection of validator for spectrum object"""

    @staticmethod
    def check_members(spectra: List[SpectrumDataLoader]):
        """checks member of lists"""
        _false_spectra = [
            spec
            for spec in spectra
            if type(spec) != SpectrumDataLoader or not hasattr(spec, "clean_data")
        ]
        if _false_spectra:
            logger.warning(
                f'_check_members not all spectra members are "SpectrumDataLoader" or missing clean_data attribute'
            )

    @staticmethod
    def check_spectra_lengths(spectra: List[SpectrumDataLoader]) -> List:
        lengths = [i.spectrum_length for i in spectra]
        set_lengths = set(lengths)
        if len(set_lengths) == 1:
            #  print(f'Spectra all same length {set_lengths}')
            pass
        else:
            length_counts = [(i, lengths.count(i)) for i in set_lengths]
            best_guess_length = max(length_counts, key=itemgetter(1))[0]
            print(f"Spectra not same length {length_counts} took {best_guess_length}")
            # self._raw_spectra = self._spectra
            spectra = [
                spec for spec in spectra if spec.spectrum_length == best_guess_length
            ]
        return spectra



import logging
import unittest
import warnings

from raman_fitting.utils.coordinators import FieldsTracker, FieldsTrackerWarning

logger = logging.getLogger(__name__)
logging.captureWarnings(True)  # sends these warning to the logger


def ignore_warnings(test_func):
    def do_test(self, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            test_func(self, *args, **kwargs)

    return do_test


class TestFieldsTracker(unittest.TestCase):
    @ignore_warnings
    def testFCO(self):

        #%%
        fco = FieldsTracker(
            fields=["peak_name", "peak_type", "param_hints"],
            sources=("kwargs", "cls_dict", "init"),
        )
        assert not fco.results
        assert fco.status == False

        fco.store("kwargs", "peak_name", "R2D2")

        with self.assertWarns(UserWarning) as cm:
            fco.store("cls_dict", "peak_name", "R4D4")
        self.assertIn(
            "Field peak_name has multiple sources", ", ".join(map(str, cm.warning.args))
        )

        with self.assertWarns(UserWarning) as cm:
            fco.store("notinsources", "peak_name", "R4D4")
        self.assertIn(
            "Store in notinsources at peak_name", ", ".join(map(str, cm.warning.args))
        )

        fco.store("init", "peak_type", "Voirentzian")

        with self.assertWarns(UserWarning) as cm:
            fco.store("init", "peak_type", "Voirentzian")
        self.assertIn(
            "Redefinition of peak_type in init ignored",
            ", ".join(map(str, cm.warning.args)),
        )

        fco.store("init", "peak_name", "Voirentzian")
        fco.store("init", "param_hints", "Voirentzian")
        fco.store("init", "peak_type", "Voirentzian")

        assert set(fco.results.keys()) == set(["peak_name", "peak_type", "param_hints"])
        assert fco.status == True

        _test_dict = {
            "peak_name": "multi_store",
            "peak_type": "multi_test",
            "param_hints": 2,
        }
        fco.multi_store("cls_dict", **_test_dict)
        assert fco.register["cls_dict"] == _test_dict
        assert fco.results["param_hints"]["value"] == 2



# pylint: disable=W0614,W0401,W0611,W0622,C0103,E0401,E0402
import logging
import sys
from itertools import repeat, starmap
from pathlib import Path

import pandas as pd

import raman_fitting
from raman_fitting.config import filepath_settings
from raman_fitting.config.filepath_helper import get_directory_paths_for_run_mode
from raman_fitting.deconvolution_models.fit_models import Fitter
from raman_fitting.deconvolution_models.base_model import InitializeModels
from raman_fitting.exporting.exporter import Exporter
from raman_fitting.indexing.indexer import MakeRamanFilesIndex
from raman_fitting.processing.spectrum_constructor import (
    SpectrumDataCollection,
    SpectrumDataLoader,
)

# from processing.cleaner import SpectrumCleaner
from raman_fitting.processing.spectrum_template import SpectrumTemplate
from raman_fitting.interfaces.cli import RUN_MODES

if __name__ == "__main__":
    pass


# else:
#     # from .. import __package_name__
#     from ..cli.cli import RUN_MODES
#     from ..config import config
#     from ..config.filepaths import get_directory_paths_for_run_mode
#     from ..deconvolution_models.fit_models import Fitter
#     from ..deconvolution_models.base_model import InitializeModels
#     from ..export.exporter import Exporter
#     from ..indexer.indexer import MakeRamanFilesIndex
#     from ..processing.spectrum_constructor import (
#         SpectrumDataCollection,
#         SpectrumDataLoader,
#     )

#     # from processing.cleaner import SpectrumCleaner
#     from ..processing.spectrum_template import SpectrumTemplate

logger = logging.getLogger(__name__)
# from raman_fitting.indexer.indexer import OrganizeRamanFiles
# from raman_fitting.processing.spectrum_constructor import SpectrumDataLoader, SpectrumDataCollection


class prdError(Exception):
    """Base error raised by pyramdeconv (future package name of raman_fitting)."""


class MainDelegatorError(prdError):
    """Raised when a method in the main delegator fails."""


class MainDelegator:
    # IDEA Add flexible input handling for the cli, such a path to dir, or list of files
    #  or create index when no kwargs are given.
    """
    Main delegator for the processing of files containing Raman spectra.

    Input parameters is DataFrame of index
    Creates plots and files in the config RESULTS directory.
    """

    # DEFAULT_RUN_MODE = {'run_mode' : 'normal'}
    # _kwargs = {}

    def __init__(self, run_mode="normal", **kwargs):

        # print(f'{self} kwargs:', kwargs)
        self.kwargs = kwargs
        self._cqnm = __class__.__qualname__

        self.run_mode = run_mode
        if run_mode not in RUN_MODES:
            logger.warning(
                f"{self}\n\twarning run_mode choice {run_mode} not in\n\t{RUN_MODES}"
            )

        self.dest_dirs = get_directory_paths_for_run_mode(run_mode=run_mode)
        self.RESULTS_DIR = self.dest_dirs["RESULTS_DIR"]
        self.DATASET_DIR = self.dest_dirs["DATASET_DIR"]
        self.INDEX_FILE = self.dest_dirs["INDEX_FILE"]

        self.spectrum = SpectrumTemplate()
        # self.index = RamanIndex

        self.run_delegator(**self.kwargs)

    def index_delegator(self, **kwargs):

        RF_index = MakeRamanFilesIndex(
            **kwargs,
        )
        # run_mode=self.run_mode,
        # RESULTS_DIR=self.RESULTS_DIR,
        # DATASET_DIR=self.DATASET_DIR,
        # INDEX_FILE=self.INDEX_FILE,
        logger.info(f"index_delegator index prepared with len {len(RF_index)}")
        return RF_index

    def run_delegator(self, **kwargs):
        # IDEA remove self.set_models() removed InitModels
        self._failed_samples = []
        self.export_collect = []

        # assert type(self.index) == type(pd.DataFrame())
        if self.run_mode in ("normal", "DEBUG", "make_index", "make_examples"):
            RF_indexer = self.index_delegator(
                run_mode=self.run_mode, dataset_dirs=self.dest_dirs, **kwargs
            )
            self.index = RF_indexer.index_selection

            if self.index.empty:
                logger.warning(f"{self._cqnm} index selection empty")

            if self.run_mode == "make_index":
                logger.info(
                    f"{self._cqnm} Debug run mode {self}. Index loaded {RF_indexer}"
                )
                sys.exit(0)

            models = self.initialize_default_models()
            self.kwargs.update({"models": models})
            # IDEA built in a model selection keyword, here or at fitting level and for the cli
            logger.info(
                f"\n{self._cqnm} models initialized for run mode ({self.run_mode}):\n\n{repr(models)}"
            )

            if self.run_mode in ("normal", "make_examples"):
                if not self.index.empty:
                    logger.debug(f"{self._cqnm}. starting run generator.")
                    # self.index

                    self._run_gen(**self.kwargs)
                else:
                    pass
                # info raman loop finished because index is empty
            elif self.run_mode == "DEBUG":
                logger.debug(f"Debug run mode {self}. Models initialized {models}")

                try:
                    # self._run_gen() # IDEA add extra test runs in tests dir
                    pass
                except Exception as e:
                    raise MainDelegatorError(
                        "The debug run failed. " f" on {self} because {e}"
                    )
                # raise('Error in DEBUG run: ', e)
            else:
                logger.warning(f"Debug run mode {self.run_mode} not recognized")
            # IDEA get testing from index and run
        else:
            logger.warning(f'Debug run mode "{self.run_mode}" not recognized not in ')
            # warning run mode not recognized

    def initialize_default_models(self):
        try:
            return InitializeModels()
        except Exception as e:
            raise MainDelegatorError(
                "The initialization of models failed. " f" on {self} with excp: {e}"
            )
            return None

    def sample_group_gen(self):
        """Generator for Sample Groups, yields the name of group and group of the index SampleGroup"""
        for grpnm, sGrp_grp in self.index.groupby(
            self.spectrum.grp_names.sGrp_cols[0]
        ):  # Loop over SampleGroups
            yield grpnm, sGrp_grp

    def _sID_gen(self, grpnm, sGrp_grp):
        """Generator for SampleIDs, yields the name of group, name of SampleID and group of the index of the SampleID"""
        for nm, sID_grp in sGrp_grp.groupby(
            list(self.spectrum.grp_names.sGrp_cols[1:])
        ):  # Loop over SampleIDs within SampleGroup
            yield (grpnm, nm, sID_grp)

    def _run_gen(self, **kwargs):
        # #IDEA sort of coordinator coroutine, can implement still deque
        sgrp_grpby = self.index.groupby(self.spectrum.grp_names.sGrp_cols[0])
        logger.info(f"{self._cqnm} _run_gen starting: {kwargs}")
        _mygen = self._generator(sgrp_grpby, **kwargs)
        Exporter(self.export_collect)  # clean up and 
        logger.info(
                    f"\n{self._cqnm} run finished.\n Results saved in {self.RESULTS_DIR}"
                )

    def _generator(self, sgrp_grpby, **kwargs):
        # _sgrpgen = self.sample_group_gen()
        export_collect = []
        for sgrpnm, sgrp_grp in sgrp_grpby:
            sID_grpby = sgrp_grp.groupby(list(self.spectrum.grp_names.sGrp_cols[1:]))
            # _sID_gen = self._sID_gen(grpnm, sgrp_grp)
            logger.info(f"{self._cqnm} _generator starting group: {sgrpnm}")
            exporter_sample = None
            for sIDnm, sIDgrp in sID_grpby:
                try:
                    exporter_sample = self.simple_process_sample_wrapper(
                        sgrpnm, sIDnm, sIDgrp, **kwargs
                    )
                    # yield from self.simple_process_sample_wrapper(_sID_gen, **kwargs)
                    # starmap(process_sample_wrapper, args_for_starmap )
                    # args_for_starmap = zip(repeat(self.process_sample), _sID_gen, repeat(kwargs))
                except GeneratorExit:
                    logger.warning(f"{self._cqnm} _generator closed.")
                    return ()
                except Exception as e:
                    logger.warning(f"{self._cqnm} _generator exception: {e}")
                    # return ()
                export_collect.append(exporter_sample)
        return export_collect

    def coordinator(self):
        pass

    def simple_process_sample_wrapper(self, *sID_args, **kwargs):
        # if _gen:
        # for sID_args in _gen:
        logger.info(
            f"{self._cqnm} starting simple process_sample_wrapper args:\n\t - {sID_args[0]}\n\t - {kwargs.keys()}"
        )
        exporter_sample = None
        try:
            logger.debug(
                f"{self._cqnm} simple process_sample_wrapper starting:\n\t - {sID_args[1]}"
            )
            exporter_sample = self.process_sample(*sID_args, **kwargs)
            if exporter_sample:
                logger.debug(
                    f"{self._cqnm} simple process_sample_wrapper appending export:\n\t - {exporter_sample}"
                )
                self.export_collect.append(exporter_sample)
        # except StopIteration:
        # logger.info(
        #     f"{self._cqnm} _gen StopIteration for simple process wrapper "
        # )
        except Exception as e:
            logger.warning(
                f"{self._cqnm} simple process_sample_wrapper exception on call process sample: {e}"
            )
            self._failed_samples.append((e, sID_args, kwargs))
        return exporter_sample

    def test_positions(
        self, sGrp_grp, sIDnm, grp_cols=["FileStem", "SamplePos", "FilePath"]
    ):
        if sGrp_grp.FileStem.nunique() != sGrp_grp.SamplePos.nunique():
            logger.warning(
                f"{sGrp_grp[grp_cols]} Unique files and positions not matching for {sIDnm}"
            )
        return sGrp_grp.groupby(grp_cols), grp_cols

    def process_sample(self, sgrpnm, sIDnm, sID_grp, **kwargs):
        """
        Loops over individual Sample positions (files) from a SampleID and performs the
        fitting, plotting and exporting.
        """
        logger.info(
            f"{self._cqnm} process_sample called:\n\t - {sgrpnm}, {sIDnm}\n\t - {kwargs.keys()}"
        )
        # self = args[0]
        # args = args[]
        # grpnm, nm, sID_grp = args

        models = kwargs.get("models", None)

        # sGr, (sID, sDate) = sgrpnm, nm # just for developer

        sGr_out = dict(zip(self.spectrum.grp_names.sGrp_cols, (sgrpnm,) + sIDnm))
        export_info_out = add_make_sample_group_destdirs(sID_grp)
        sample_pos_grp, sPos_cols = self.test_positions(
            sID_grp, sIDnm, list(self.spectrum.grp_names.sPos_cols)
        )

        sample_spectra = []
        for meannm, meangrp in sample_pos_grp:
            logger.info(f"{self._cqnm} process sample mean loop file: {meannm}.")
            sPos_out = dict(zip(self.spectrum.grp_names.sPos_cols, meannm))
            _spectrum_position_info_kwargs = {**sGr_out, **export_info_out, **sPos_out}
            spectrum_data = SpectrumDataLoader(
                file=meannm[-1], run_kwargs=_spectrum_position_info_kwargs, ovv=meangrp
            )
            # spectrum_data.plot_raw()
            sample_spectra.append(spectrum_data)
        if sample_spectra:
            spectra_collection = SpectrumDataCollection(sample_spectra)
            ft = Fitter(spectra_collection, RamanModels=models)
            rex = Exporter(ft)
            return rex
        else:
            logger.info(
                f"{self._cqnm} process sample spectra empty {','.join(map(str,[sgrpnm, sIDnm]))}."
            )
            return None

    def __repr__(self):
        return f'Maindelegator: run_mode = {self.run_mode}, {", ".join([f"{k} = {str(val)}" for k,val in self.kwargs.items()])}'

        # _count = 0
        # while True:
        #     try:
        #         next(_mygen)
        #         _count += 1
        #     except StopIteration:
        #         logger.debug(
        #             f"{self._cqnm} _run_gen StopIteration after {_count } steps"
        #         )
        #         # print('StopIteration for mygen')
        #         break
        #     finally:


def add_make_sample_group_destdirs(sample_grp: pd.DataFrame):

    dest_grp_dir = Path(
        sample_grp.DestDir.unique()[0]
    )  # takes one destination directory from Sample Groups
    dest_fit_plots = dest_grp_dir.joinpath("Fitting_Plots")
    dest_fit_comps = dest_grp_dir.joinpath("Fitting_Components")
    dest_fit_comps.mkdir(parents=True, exist_ok=True)

    dest_raw_data_dir = dest_grp_dir.joinpath("Raw_Data")
    dest_raw_data_dir.mkdir(parents=True, exist_ok=True)

    export_info = {
        "DestGrpDir": dest_grp_dir,
        "DestFittingPlots": dest_fit_plots,
        "DestFittingComps": dest_fit_comps,
        "DestRaw": dest_raw_data_dir,
    }
    return export_info


def process_sample_wrapper(fn, *args, **kwargs):
    def wrapper(*args, **kwargs):
        logger.debug(
            f"process_sample_wrapper args:\n\t - {fn}\n\t - {args}\n\t - {kwargs.keys()}"
        )
        exp_sample = None
        try:
            exp_sample = fn(*args, **kwargs)
            # self.export_collect.append(exp_sample)
        except Exception as e:
            logger.error(
                f"process_sample_wrapper process_sample_wrapper exception on call {fn}: {e}"
            )
            # self._failed_samples.append((e, args, kwargs))
            exp_sample = (e, args, kwargs)

        return exp_sample


def make_examples():
    _main_run = MainDelegator(run_mode="make_examples")
    return _main_run



class prdError(Exception):
    """Base error raised by pyramdeconv."""


class MainDelegatorError(prdError):
    """Raised when a method in the main delegator fails."""


class DataBaseError(prdError):
    """Raised when interaction with the default database fails"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 14:49:50 2020

@author: DW
"""
import matplotlib
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.ticker import AutoMinorLocator, FormatStrFormatter, MultipleLocator

matplotlib.rcParams.update({"font.size": 14})

#%%

# IDEA PLOTTING PER PEAK MODEL
def plotting_info(windowname):  # pragma: no cover
    axes = {
        "full": (0, 0),
        "low": (0, 1),
        "1st_order": (0, 2),
        "mid": (1, 1),
        "2nd_order": (1, 2),
        "normalization": (1, 0),
    }
    return axes.get(windowname)


def raw_data_export(fitting_specs):  # pragma: no cover
    current_sample = fitting_specs[0].windowname, fitting_specs[0].sIDmean_col
    try:
        raw_data_spectra_plot(fitting_specs)
    except Exception as e:
        print("no extra Raw Data plots for {1} \n {0}".format(e, current_sample))
    try:
        raw_data_spectra_export(fitting_specs)
    except Exception as e:
        print("no extra Raw Data plots for {1} \n {0}".format(e, current_sample))


def raw_data_spectra_plot(fitting_specs):  # pragma: no cover
    #    fitting_specs
    try:
        fig, ax = plt.subplots(2, 3, figsize=(18, 12))
        ax_wn = []

        for spec in fitting_specs:
            try:
                ax_wn = ax[plotting_info(spec.windowname)]
                #                spec.windowname
                _legend = True if "full" == spec.windowname else False
                spec.mean_spec.plot(
                    x="ramanshift",
                    y=spec.sID_rawcols,
                    ax=ax_wn,
                    alpha=0.5,
                    legend=_legend,
                )
                spec.mean_spec.plot(
                    x="ramanshift",
                    y=spec.sIDmean_col,
                    ax=ax_wn,
                    c="k",
                    alpha=0.7,
                    lw=3,
                    legend=_legend,
                )

                ax_wn.set_title(spec.windowname)
                #                _legend = True if 'full' in spec.windowname else False
                if _legend:
                    ax_wn.legend(fontsize=10)

            except:
                pass

        plt.suptitle(spec.sIDmean_col, fontsize=16)
        plt.savefig(
            spec.mean_info.DestRaw.unique()[0].joinpath(f"{spec.sIDmean_col}.png"),
            dpi=300,
            bbox_inches="tight",
        )
        plt.close()
    except Exception as e:
        print("no extra Raw Data plots: {0}".format(e))


def raw_data_spectra_export(fitting_specs):
    #    fitting_specs
    try:
        for spec in fitting_specs:
            #            spec.windowname, spec.sIDmean_col
            wnxl_outpath_spectra = spec.mean_info.DestRaw.unique()[0].joinpath(
                f"spectra_{spec.sIDmean_col}_{spec.windowname}.xlsx"
            )
            spec.mean_spec.to_excel(wnxl_outpath_spectra)

        _0_spec = fitting_specs[0]
        wnxl_outpath_info = _0_spec.mean_info.DestRaw.unique()[0].joinpath(
            f"info_{_0_spec.sIDmean_col}.xlsx"
        )
        _0_spec.mean_info.to_excel(wnxl_outpath_info)

    #            ax_wn = ax[plotting_info(spec.windowname)]
    except Exception as e:
        print("no extra Raw Data plots: {0}".format(e))


def fit_spectrum_plot(
    peak1,
    peak2,
    res1_peak_spec,
    res2_peak_spec,
    plot_Annotation=True,
    plot_Residuals=True,
):  # pragma: no cover

    modname_2 = peak2
    #%%
    sID = res1_peak_spec.extrainfo["SampleID"]
    SampleBgmean_col = res1_peak_spec.raw_data_col

    #    sID, DestPlotDir, SampleBgmean_col, FitData, FitModPeaks, FitData_2nd,comps, comps_2nd,out, out_2nd,
    FitData_1st = res1_peak_spec.FitComponents
    Model_peak_col_1st = res1_peak_spec.model_name
    Model_data_col_1st = res1_peak_spec.model_name
    # f'Model_{Model_peak_col_1st}'
    compscols_1st = [
        i for i in FitData_1st.columns if i.endswith("_") and not i.startswith("Si")
    ]
    #    FitReport_1st = res1_peak_spec.FitReport

    FitData_2nd = res2_peak_spec.FitComponents
    Model_peak_col_2nd = res2_peak_spec.model_name
    Model_data_col_2nd = res2_peak_spec.model_name
    # f'Model_{Model_peak_col_2nd}'
    compscols_2nd = [i for i in FitData_2nd.columns if i.endswith("_")]

    FitPars, FitPars_2nd = res1_peak_spec.FitParameters, res2_peak_spec.FitParameters

    fig = plt.figure(figsize=(28, 24))
    gs = gridspec.GridSpec(4, 1, height_ratios=[4, 1, 4, 1])
    ax = plt.subplot(gs[0])
    axRes = plt.subplot(gs[1])
    #                axAnn = plt.subplot(gs[1])
    ax2nd = plt.subplot(gs[2])
    ax2ndRes = plt.subplot(gs[3])
    ax2ndRes.grid(True), axRes.grid(True, "both")
    ax2nd.grid(True), ax.grid(True, "both")
    ax.get_yaxis().set_tick_params(direction="in")
    ax.get_xaxis().set_tick_params(direction="in")
    ax.set_title(SampleBgmean_col)
    #                             '%s' %FileName)
    ax.xaxis.set_minor_locator(AutoMinorLocator(2))
    ax.yaxis.set_minor_locator(AutoMinorLocator(2))
    ax.tick_params(which="both", direction="in")
    ax2nd.xaxis.set_minor_locator(AutoMinorLocator(2))
    ax2nd.yaxis.set_minor_locator(AutoMinorLocator(2))
    ax2nd.tick_params(which="both", direction="in")
    ax.set_facecolor("oldlace"), ax2nd.set_facecolor("oldlace")
    axRes.set_facecolor("oldlace"), ax2ndRes.set_facecolor("oldlace")
    #                for FitN in [PosInts_Bg,PosInts_Bg_2nd]:
    #                    fig,ax = plt.subplots(1,1,figsize=(12,12))
    #                if FitN.RamanShift.max() > 2300:
    ax2nd.plot(
        FitData_2nd["RamanShift"],
        FitData_2nd[Model_data_col_2nd],
        label=Model_data_col_2nd,
        lw=3,
        c="r",
    )
    ax2nd.plot(
        FitData_2nd["RamanShift"],
        FitData_2nd[res2_peak_spec.raw_data_col],
        label="Data",
        lw=3,
        c="grey",
        alpha=0.5,
    )
    if plot_Residuals:
        ax2ndRes.plot(
            FitData_2nd["RamanShift"],
            FitData_2nd[res2_peak_spec.raw_data_col] - FitData_2nd[Model_data_col_2nd],
            label="Residual",
            lw=3,
            c="k",
            alpha=0.8,
        )

    for fit_comp_col_2nd in compscols_2nd:  # automatic color cycle 'cyan' ...
        ax2nd.plot(
            FitData_2nd["RamanShift"],
            FitData_2nd[fit_comp_col_2nd],
            ls="--",
            lw=4,
            label=fit_comp_col_2nd,
        )
        center_col, height_col = (
            fit_comp_col_2nd + "center",
            fit_comp_col_2nd + "height",
        )
        ax2nd.annotate(
            f"{fit_comp_col_2nd}\n {FitPars_2nd[center_col].round(0).iloc[0]:.0f}",
            xy=(
                FitPars_2nd[center_col].iloc[0] * 0.97,
                0.8 * FitPars_2nd[height_col].iloc[0],
            ),
            xycoords="data",
        )
    #    ax2nd.plot(FitData_2nd['RamanShift'], comps_2nd['D1D1_'], color='lime',ls='--',lw=4,label='2*D')
    #    ax2nd.plot(FitData_2nd['RamanShift'], comps_2nd['D4D4_'], color='purple',ls='--',lw=4,label='2*D4')
    #    ax2nd.plot(FitData_2nd['RamanShift'], comps_2nd['GD1_'], color='pink',ls='--',lw=4,label='G+D1')
    ax2nd.set_ylim(-0.02, FitData_2nd[Model_data_col_2nd].max() * 1.5)
    #                else:
    ax.plot(
        FitData_1st["RamanShift"],
        FitData_1st[Model_data_col_1st],
        label=Model_data_col_1st,
        lw=3,
        c="r",
    )
    ax.plot(
        FitData_1st["RamanShift"],
        FitData_1st[res1_peak_spec.raw_data_col],
        label="Data",
        lw=3,
        c="grey",
        alpha=0.8,
    )
    
    if plot_Residuals:
        axRes.plot(
            FitData_1st["RamanShift"],
            FitData_1st[res1_peak_spec.raw_data_col] - FitData_1st[Model_data_col_1st],
            label="Residual",
            lw=3,
            c="k",
            alpha=0.8,
        )

    for fit_comp_col_1st in compscols_1st:  # automatic color cycle 'cyan' ...
        ax.plot(
            FitData_1st["RamanShift"],
            FitData_1st[fit_comp_col_1st],
            ls="--",
            lw=4,
            label=fit_comp_col_1st,
        )
        center_col, height_col = (
            fit_comp_col_1st + "center",
            fit_comp_col_1st + "height",
        )
        ax.annotate(
            f"{fit_comp_col_1st}:\n {FitPars[center_col].round(0).iloc[0]:.0f}",
            xy=(FitPars[center_col].iloc[0] * 0.97, 0.7 * FitPars[height_col].iloc[0]),
            xycoords="data",
        )

    #    ax.plot(FitData_1st['RamanShift'], FitData_1st['D_'], color='lime',ls='--',lw=4,label='D')
    #                ax.plot(FitData['RamanShift'], comps['D2_'], color='grey',ls='--',lw=4,label='D2')
    #    ax.plot(FitData_1st['RamanShift'], FitData_1st['D3_'], 'b--',lw=4,label='D3')
    #    ax.plot(FitData_1st['RamanShift'], FitData_1st['D4_'], color='purple',ls='--',lw=4,label='D4')
    if "peaks" in peak1:
        #        ax.plot(FitData_1st['RamanShift'], FitData_1st['D2_'], color='magenta',ls='--',lw=4,label='D2')
        #        ax.annotate('D2:\n %.0f'%FitPars['D2_center'],xy=(FitPars['D2_center']*0.97,0.8*FitPars['I_D2']),xycoords='data')
        if peak1.endswith("+Si"):
            ax.plot(
                FitData_1st["RamanShift"],
                FitData_1st["Si1_"],
                "b--",
                lw=4,
                label="Si_substrate",
            )
            if FitPars["Si1_fwhm"].iloc[0] > 1:
                ax.annotate(
                    "Si_substrate:\n %.0f" % FitPars["Si1_center"],
                    xy=(FitPars["Si1_center"] * 0.97, 0.8 * FitPars["Si1_height"]),
                    xycoords="data",
                )
    #        if '6peaks' in FitModPeaks:
    #            ax.plot(FitData_1st['RamanShift'], FitData_1st['D5_'],  color='darkorange',ls='--',lw=4,label='D5')
    #            ax.annotate('D5:\n %.0f'%FitPars['D5_center'],xy=(FitPars['D5_center']*0.97,0.8*FitPars['I_D5']),xycoords='data')
    ##                    for colN,col in FitN.iteritems():
    #                        if 'RamanShift' in colN or colN.split('_')[-1] == 'count' or colN.split('_')[-1] == 'std' :
    #                            continue
    #                        w,i = FitN['RamanShift'].values,col.values
    #                        if 'mean' in colN:
    #                            ax.plot(w[::],i[::],label=colN)
    #                        else:
    #                            ax.scatter(w[::],i[::],label=colN)
    if plot_Annotation:
        frsplit = res1_peak_spec.FitReport.split()
        if len(frsplit) > 200:
            fr = res1_peak_spec.FitReport.replace("prefix='D3_'", "prefix='D3_' \n")
        else:
            fr = res1_peak_spec.FitReport
        props = dict(boxstyle="round", facecolor="wheat", alpha=0.5)
        Report1 = ax.text(
            1.01,
            1,
            fr,
            transform=ax.transAxes,
            fontsize=11,
            verticalalignment="top",
            bbox=props,
        )
        Report2 = ax2nd.text(
            1.01,
            0.7,
            res2_peak_spec.FitReport,
            transform=ax2nd.transAxes,
            fontsize=11,
            verticalalignment="top",
            bbox=props,
        )

    ax.legend(loc=1), ax.set_xlabel("Raman shift (cm$^{-1}$)"), ax.set_ylabel(
        "normalized I / a.u."
    )
    ax2nd.legend(loc=1), ax2nd.set_xlabel("Raman shift (cm$^{-1}$)"), ax2nd.set_ylabel(
        "normalized I / a.u."
    )
    #                plt.show()

    #    plt.show()
    plt.savefig(
        res1_peak_spec.extrainfo["DestFittingModel"].with_suffix(".png"),
        dpi=100,
        bbox_extra_artists=(Report1, Report2),
        bbox_inches="tight",
    )
    plt.close()
    #%%



import datetime
import unittest

# from raman_fitting.deconvolution_models import first_order_peaks
import pandas as pd
import pytest

import raman_fitting
from raman_fitting.datafiles import example_files
from raman_fitting.deconvolution_models.base_model import InitializeModels
from raman_fitting.delegating.main_delegator import (
    MainDelegator,
    add_make_sample_group_destdirs,
)


class TestMainDelegator(unittest.TestCase):
    def setUp(self):
        self.maindebug = MainDelegator(run_mode="DEBUG")
        self.models = self.maindebug.initialize_default_models()

    def test_initialize_models(self):
        self.assertTrue(isinstance(self.models, InitializeModels))

    def test_index(self):
        self.assertTrue(hasattr(self.maindebug, "index"))
        self.assertTrue(isinstance(getattr(self.maindebug, "index"), pd.DataFrame))

        _sample_group = self.maindebug.sample_group_gen()
        _arg = next(_sample_group)
        _destdirs = add_make_sample_group_destdirs(_arg[-1])
        _alltest = all([_arg[0] in a for a in [i.parts for i in _destdirs.values()]])
        self.assertTrue(_alltest)

    def test_generator(self):

        _sample_group = self.maindebug.sample_group_gen()
        _sample_group_arg = next(_sample_group)
        self.assertTrue(_sample_group_arg)

        _sID_gen = self.maindebug._sID_gen(*_sample_group_arg)
        _sID_arg = next(_sID_gen)
        self.assertTrue(_sID_arg)

        # while True:

        #     try:

        #     except StopIteration as e:
        #         print(e)


if __name__ == "__main__":
    unittest.main()
    self = TestMainDelegator()



""" Peaks used for normalization"""

if __name__ == "__main__":
    from raman_fitting.deconvolution_models.base_peak import BasePeak
else:
    from .base_peak import BasePeak

__all__ = ["norm_G_peak", "norm_D_peak"]
# ====== PEAKS USED FOR NORMALIZATION ======= #


class norm_G_peak(metaclass=BasePeak):
    """G_peak used for normalization"""

    def __init__(self, *args, **kwargs):
        self.peak_name = "norm_G"
        self.peak_type = "Lorentzian"
        self.input_param_settings = {
            "center": {"value": 1581, "min": 1500, "max": 1600},
            "sigma": {"value": 40, "min": 1e-05, "max": 1e3},
            "amplitude": {"value": 8e4, "min": 1e2},
        }

    # norm_G_peak = BasePeak(peak_name=peak_name,peak_type=peak_type,input_param_settings=settings)


class norm_D_peak(metaclass=BasePeak):
    """D_peak for normalization"""

    def __init__(self, *args, **kwargs):
        self.peak_name = "norm_D"
        self.peak_type = "Lorentzian"
        self.input_param_settings = {
            "center": {"value": 1350, "min": 1300, "max": 1400},
            "sigma": {"value": 90, "min": 1e-05},
            "amplitude": {"value": 10e5, "min": 1e2},
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy

import numpy as np
import pandas as pd
from scipy import signal
from scipy.stats import linregress

# from collections import namedtuple
from raman_fitting.processing.spectrum_template import (
    SpecTemplate,
    SpectrumWindowLimits,
    SpectrumWindows,
)

# if __name__ == "__main__":
#     pass
# else:
#     from .spectrum_template import SpecTemplate, SpectrumWindowLimits, SpectrumWindows


class SpectrumMethodException(ValueError):
    pass


class SpectrumMethods:
    """
    Parent class to hold several Spetrum Methods as children
    """

    data = SpecTemplate

    def __init__(self, ramanshift, intensity, label="", **kwargs):
        """


        Parameters
        ----------
        ramanshift : array or list
            collection of the ramanshift values
        intensity : array or list
            collection of the intensity values
        label : TYPE, optional
            DESCRIPTION. The default is "".
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.ramanshift = ramanshift
        self.intensity = intensity
        self.label = label
        self.kwargs = kwargs

    @staticmethod
    def filtered_int(intensity=None):

        try:
            int_savgol_fltr = signal.savgol_filter(intensity, 13, 3, mode="nearest")
        except Exception as e:
            raise SpectrumMethodException(f"no intensity given to filter, {e}")
            int_savgol_fltr = []
        return int_savgol_fltr


class SpectrumSplitter(SpectrumMethods):
    """
    For splitting of spectra into the several SpectrumWindows,
    the names of the windows are taken from SpectrumWindows
    and set as attributes to the instance.
    """

    def __init__(self, *args, **kws):
        super().__init__(*args, **kws)

    def split_data(self, spec_windows=SpectrumWindows()):
        _r, _int = self.ramanshift, self.intensity
        # self.register.get(on_lbl)
        _d = {}
        for windowname, limits in spec_windows.items():
            ind = (_r >= np.min(limits)) & (_r <= np.max(limits))
            # _intslice = _int[ind]
            label = f"window_{windowname}"
            if self.label:
                label = f"{self.label}_{label}"
            _data = self.data(_r[ind], _int[ind], label)
            setattr(self, label, _data)
            _d.update(**{windowname: _data})
        self.windows_data = _d


# class Filter(SpectrumMethods):
# '''
# For filtering the intensity of a spectrum
# '''

# def get_filtered(self, intensity):


class BaselineSubtractorNormalizer(SpectrumSplitter):
    """
    For baseline subtraction as well as normalization of a spectrum
    """

    def __init__(self, *args, **kws):
        super().__init__(*args, **kws)
        self.split_data()
        self.windowlimits = SpectrumWindowLimits()
        self.subtract_loop()
        self.get_normalization_factor()
        self.set_normalized_data()

    def subtract_loop(self):
        _blcorr = {}
        _info = {}
        for windowname, spec in self.windows_data.items():
            blcorr_int, blcorr_lin = self.subtract_baseline_per_window(windowname, spec)
            label = f"blcorr_{windowname}"
            if self.label:
                label = f"{self.label}_{label}"
            _data = self.data(spec.ramanshift, blcorr_int, label)
            _blcorr.update(**{windowname: _data})
            _info.update(**{windowname: blcorr_lin})
        self.blcorr_data = _blcorr
        self.blcorr_info = _info

    def subtract_baseline_per_window(self, windowname, spec):
        rs = spec.ramanshift
        if windowname[0:4] in ("full", "norm"):
            i_fltrd_dspkd_fit = self.windows_data.get("1st_order").intensity
        else:
            i_fltrd_dspkd_fit = spec.intensity
        _limits = self.windowlimits.get(windowname)
        # assert bool(_limits),f'no limits for {windowname}'
        bl_linear = linregress(
            rs[[0, -1]],
            [
                np.mean(i_fltrd_dspkd_fit[0 : _limits[0]]),
                np.mean(i_fltrd_dspkd_fit[_limits[1] : :]),
            ],
        )
        i_blcor = spec.intensity - (bl_linear[0] * rs + bl_linear[1])
        #        blcor = pd.DataFrame({'Raman-shift' : w, 'I_bl_corrected' :i_blcor, 'I_raw_data' : i})
        return i_blcor, bl_linear

    def get_normalization_factor(self, norm_method="simple"):
        if norm_method == "simple":
            normalization_intensity = np.nanmax(
                self.blcorr_data["normalization"].intensity
            )

        if 0:
            # IDEA not implemented
            if norm_method == "fit":
                normalization = NormalizeFit(
                    self.blcorr_data["1st_order"], plotprint=False
                )  # IDEA still implement this NormalizeFit
                normalization_intensity = normalization["IG"]

        self.norm_factor = 1 / normalization_intensity
        norm_dict = {
            "norm_factor": self.norm_factor,
            "norm_method": norm_method,
            "norm_int": normalization_intensity,
        }
        self.norm_dict = norm_dict

    def set_normalized_data(self):
        _normd = {}
        for windowname, spec in self.blcorr_data.items():

            label = f"norm_blcorr_{windowname}"
            if self.label:
                label = f"{self.label}_{label}"

            _data = self.data(spec.ramanshift, spec.intensity * self.norm_factor, label)
            _normd.update(**{windowname: _data})
        self.norm_data = _normd


def NormalizeFit(spec, plotprint=False):
    pass  # IDEA placeholder


def array_nan_checker(array):
    _nans = [n for n, i in enumerate(array) if np.isnan(i)]
    return _nans


class Despiker(SpectrumMethods):
    """
    A Despiking algorithm from reference literature: https://doi.org/10.1016/j.chemolab.2018.06.009

    Parameters
    ----------
    input_intensity : np.ndarray
        The intensity array of which the desipked intensity will be calculated.
    info : dict, optional
        Extra information for despiking settings are added to this dict.
    Attributes
    ---------
    despiked_intensity : np.ndarray
        The resulting array of the despiked intensity of same length as input_intensity.
    Notes
    --------
    Let Y1;...;Yn represent the values of a single Raman spectrum recorded at equally spaced wavenumbers.
    From this series, form the detrended differenced seriesr Yt ...:This simple
    ata processing step has the effect of annihilating linear and slow movingcurve linear trends, however,
    sharp localised spikes will be preserved.Denote the median and the median absolute deviation of
    D.A. Whitaker, K. Hayes. Chemometrics and Intelligent Laboratory Systems 179 (2018) 82–84
    """

    keys = ["input_intensity", "Zt", "Z_t_filtered", "despiked_intensity"]

    def __init__(
        self, intensity: np.array([]), Z_threshold=4, moving_window_size=1, info={}
    ):
        self.info = info
        self.Z_threshold = Z_threshold
        self.moving_window_size = moving_window_size

        self.info.update(
            {"Z_threshold": Z_threshold, "Z_filter_ma": moving_window_size}
        )

        # these get populated by the run_despike call in the setter
        self.result = {}
        self.df = pd.DataFrame()

        # setter calls to run_despike
        self._int = intensity

        # _new_int = copy.deepcopy(intensity)
        self.input_intensity = intensity

    @property
    def input_intensity(self):
        return self._input_intensity

    @input_intensity.setter
    def input_intensity(self, value):
        """sanitizes the input argument value for an array"""

        type_test = str(type(value))
        if "__main__" in type_test:
            if "intensity" in value._fields:
                val_intensity = value.intensity
        elif "numpy.ndarray" in type_test:
            val_intensity = value
        elif "dict" in type_test:
            val_intensity = value.get([i for i in value.keys() if "intensity" in i][0])
        else:
            raise ValueError(f"Despike input error {type_test} for {value}")

        self.info.update({"input_type": type_test})
        self._input_intensity = val_intensity
        self.despiked_intensity = val_intensity

    @property
    def despiked_intensity(self):
        return self._despiked_intensity

    @despiked_intensity.setter
    def despiked_intensity(self, value):

        result = self.run_despike_steps(value, self.Z_threshold)

        self._despiked_intensity = result["despiked_intensity"]

        self.result = result
        self.df = pd.DataFrame(result)

    def run_despike_steps(self, intensity, Z_threshold):

        Z_t = self.calc_Z(intensity)
        Z_t_filtered = self.calc_Z_filtered(Z_t, Z_threshold)
        i_despiked = self.despike_filter(
            intensity, Z_t_filtered, self.moving_window_size
        )

        result_values = [intensity, Z_t, Z_t_filtered, i_despiked]
        result = dict(zip(self.keys, result_values))
        return result

    @staticmethod
    def calc_Z(intensity):
        dYt = np.append(np.diff(intensity), 0)
        #    dYt = intensity.diff()
        dYt_Median = np.median(dYt)
        #    M = dYt.median()
        #    dYt_M =  dYt-M
        dYt_MAD = np.median(abs(dYt - dYt_Median))
        #    MAD = np.mad(dYt)
        Z_t = (0.6745 * (dYt - dYt_Median)) / dYt_MAD
        #    intensity = blcor.assign(**{'abs_Z_t': Z_t.abs()})
        return Z_t

    @staticmethod
    def calc_Z_filtered(Z_t, Z_threshold):
        Z_t_filtered = copy.deepcopy(Z_t)
        Z_t_filtered[np.abs(Z_t) > Z_threshold] = np.nan
        Z_t_filtered[0] = Z_t_filtered[-1] = 0
        return Z_t_filtered
        # self.Z_t_filtered = Z_t_filtered

    #        Z_threshold = 3.5
    #        Z_t_filtered = [Z_t
    #        Z_t_filtered[Z_filter_indx] = np.nan
    #        y_out,n = intensity,len(intensity)
    @staticmethod
    def despike_filter(
        intensity, Z_t_filtered, moving_window_size, ignore_lims=(20, 46)
    ):
        n = len(intensity)
        i_despiked = copy.deepcopy(intensity)
        spikes = np.where(np.isnan(Z_t_filtered))
        for i in list(spikes[0]):

            if i < ignore_lims[0] or i > ignore_lims[1]:
                w = np.arange(
                    max(0, i - moving_window_size), min(n, i + moving_window_size)
                )
                w = w[~np.isnan(Z_t_filtered[w])]
                if intensity[w].any():
                    i_despiked[i] = np.mean(intensity[w])
                else:
                    i_despiked[i] = intensity[i]
            # else:
            # pass  # ignored
        return i_despiked

    def plot_Z(self):
        # fig,ax = plt.subplots(2)
        self.df.plot(y=["Zt", "Z_t_filtered"], alpha=0.5)
        self.df.plot(y=["input_intensity", "despiked_intensity"], alpha=0.8)
        # plt.show()
        # plt.close()



""" The member of the validated collection of BasePeaks are here assembled into fitting Models"""

import logging
from warnings import warn

from lmfit import Model

# from .. import __package_name__

logger = logging.getLogger(__name__)


_SUBSTRATE_PEAK = "Si1_peak"

from raman_fitting.deconvolution_models.peak_validation import PeakModelValidator

# if __name__ == "__main__":
#     pass
# else:
#     from .peak_validation import PeakModelValidator

#%%

# ====== InitializeMode======= #
class InitializeModels:
    """
    This class will initialize and validate the different fitting models.
    The models are of type lmfit.model.CompositeModel and stored in a dict with names
    for the models as keys.
    """

    _standard_1st_order_models = {
        "2peaks": "G+D",
        "3peaks": "G+D+D3",
        "4peaks": "G+D+D3+D4",
        "5peaks": "G+D+D2+D3+D4",
        "6peaks": "G+D+D2+D3+D4+D5",
    }
    _standard_2nd_order_models = {"2nd_4peaks": "D4D4+D1D1+GD1+D2D2"}

    def __init__(self, standard_models=True):
        self._cqnm = self.__class__.__name__

        self.peak_collection = self.get_peak_collection(PeakModelValidator)

        self.all_models = {}
        self.construct_standard_models()
        # self.normalization_model = self.peak_collection.normalization

    def get_peak_collection(self, func):
        try:
            peak_collection = func()
            logger.debug(
                f"{self._cqnm} collection of peaks validated with {func}:\n{peak_collection}"
            )

        except Exception as e:
            logger.error(f"{self._cqnm} failure in call {func}.\n\t{e}")
            peak_collection = []
        return peak_collection

    def construct_standard_models(self):

        _models = {}
        _models_1st = {
            f"1st_{key}": BaseModel(
                peak_collection=self.peak_collection, model_name=value
            )
            for key, value in self._standard_1st_order_models.items()
        }
        _models.update(_models_1st)
        _models_1st_no_substrate = {
            f"1st_{key}": BaseModel(
                peak_collection=self.peak_collection, model_name=value
            )
            for key, value in self._standard_1st_order_models.items()
        }
        _models.update(_models_1st_no_substrate)
        self.first_order = {**_models_1st, **_models_1st_no_substrate}

        _models_2nd = {
            key: BaseModel(peak_collection=self.peak_collection, model_name=value)
            for key, value in self._standard_2nd_order_models.items()
        }
        _models.update(_models_2nd)
        self.second_order = _models_2nd
        self.all_models = _models

    def __repr__(self):
        _t = "\n".join(map(str, self.all_models.values()))
        return _t


class BaseModelWarning(UserWarning):
    pass


class BaseModel:
    """
    This Model class combines the collection of valid peaks from BasePeak into a regression model of type lmfit.model.CompositeModel
    that is compatible with the lmfit Model and fit functions.
    The model_name, include_substrate and lmfit_model attributes are kept consistent w.r.t. their meaning when they are set.

    Parameters
    --------
        model_name: string ==> is converted to lmfit Model object
        include_substrate: bool ==> toggle between True and False to include a substrate peak

    """

    _SEP = "+"
    _SUFFIX = "_"

    # IDEA change include substrate to  has substrate and remove from init
    def __init__(
        self,
        model_name: str = "",
        peak_collection=PeakModelValidator(),
        substrate_peak_name: str = _SUBSTRATE_PEAK,
    ):

        self.peak_collection = peak_collection
        self.peak_options = self.set_peak_options()
        self.substrate_peak_name = substrate_peak_name
        # self.include_substrate = include_substrate
        # has_substrate: bool = False,
        self._substrate_name = self.substrate_peak_name.split(self._SUFFIX)[0]
        self.model_name = model_name

        self.lmfit_model = self.model_constructor_from_model_name(self.model_name)
        # self.model_constructor()
        # self.peak_dict = self.peak_collection.get_dict()

    def set_peak_options(self):
        _opts = {}
        for _pk in self.peak_collection.options:
            try:
                _prefix = _pk.split(self._SUFFIX)[0]
                if _prefix:
                    _opts.update({_prefix: _pk})
            except Exception as e:
                warn(
                    f'Peak {_pk} not valid name "{self._SUFFIX}, error:\n{e}',
                    BaseModelWarning,
                )
        return _opts

    @property
    def model_name(self):
        return self._model_name

    @model_name.setter
    def model_name(self, name):
        """Model name for str => model conversion"""
        _ch = True
        name = self.validate_model_name_input(name)
        if hasattr(self, "_model_name"):
            if name == self._model_name:
                _ch = False
        if _ch:
            self.lmfit_model = self.model_constructor_from_model_name(name)
            # self._equalize_name_choice(name)
            self._model_name = name

    @property
    def has_substrate(self):
        _has = False
        if hasattr(self, "model_name"):
            _has = self.name_contains_substrate(self.model_name)

        return _has

    @has_substrate.setter
    def has_substrate(self, value):
        # _hasattr_model = hasattr(self, 'model')
        raise AttributeError(
            f'{self.__class__.__name__} this property can not be set "{value}", use add_ or remove_ substrate function.'
        )
        # _ch = True
        # if hasattr(self,'_include_substrate'):
        #     if _choice == self._include_substrate:
        #         _ch = False
        # if _ch:
        #     self._equalize_name_choice(None, _choice)
        #     self._include_substrate = _choice
        # self._include_substrate = _choice

    def name_contains_substrate(self, _name):
        """Checks if name contains the substrate name, returns bool"""
        _name_contains_any = False
        if type(_name) == str:
            _name_contains_any = any(
                i == self._substrate_name for i in _name.split("+")
            )
        return _name_contains_any

    def remove_substrate(self):
        if hasattr(self, "model_name"):
            _name = self.model_name
            if self.name_contains_substrate(_name):
                warn(
                    f'\n{self.__class__.__name__} remove substrate is called so "{self._substrate_name}" is removed from {_name}.\n',
                    BaseModelWarning,
                )
                _new_name = "+".join(
                    i for i in _name.split("+") if i not in self._substrate_name
                )  # remove substr name
                if _new_name != _name:
                    self.model_name = _new_name
        # return _name

    def add_substrate(self):
        if hasattr(self, "model_name"):
            _name = self.model_name
            if not self.name_contains_substrate(_name):
                _new_name = _name + f"+{self._substrate_name}"  # add substr name
                if _new_name != _name:
                    self.model_name = _new_name
        # return _name

    def validate_model_name_input(self, value):
        """checks if given input name is valid"""
        if type(value) != str:
            raise TypeError(
                f'Given name "{value}" for model_name should be a string insteady of type({type(value).__name__})'
            )
        elif not value:
            warn(f'\n\tThis name "{value}" is an empty string', BaseModelWarning)
            return value
        elif "+" not in value:
            warn(
                f'\n\tThis name "{value}" does not contain the separator "+". (could be just 1 Peak)',
                BaseModelWarning,
            )
            return value
        else:
            _clean_string = "".join([i for i in value if i.isalnum() or i == "+"])
            _splitname = _clean_string.split("+")
            if not _splitname or not any(bool(i) for i in _splitname):
                raise ValueError(f'The split with sep "+" of name {value} is empty')
            else:
                return "+".join([i for i in _splitname if i])

    def model_constructor_from_model_name(self, _name):
        """Construct a lmfit.Model from the string model name"""

        _discarded_terms = []
        _peak_names = []
        if _name:
            for _peak in _name.split(self._SEP):  # filter model name for last time
                _peak_from_opts = self.peak_options.get(_peak, None)
                if _peak_from_opts:
                    _peak_names.append(_peak_from_opts)
                else:
                    _discarded_terms.append(_peak)

        _peak_models = [
            self.peak_collection.model_dict.get(i) for i in _peak_names if i
        ]
        if _discarded_terms:
            warn(
                f'Model evalution for "{_name}" discarded terms {",".join(_discarded_terms)} => clean: {_peak_names}',
                BaseModelWarning,
            )

        if not _peak_models:
            _lmfit_model = None
        elif len(_peak_models) == 1:
            _lmfit_model = _peak_models[0].peak_model
        elif len(_peak_models) >= 2:
            # _eval_model_name = ' + '.join([i[0] for i in _peak_models])
            _composite_model = None
            for _pkmod in _peak_models:
                _mod = _pkmod.peak_model
                if not _composite_model:
                    _composite_model = _mod
                else:
                    try:
                        _composite_model += _mod
                    except Exception as e:
                        warn(
                            f"Model add operation failed for constructing Composite Model {_pkmod.name}.\n {e}",
                            BaseModelWarning,
                        )
            _lmfit_model = _composite_model

        if not issubclass(type(_lmfit_model), Model):
            warn(
                f"Model constructor does not yield type ({type(Model)} {type(_lmfit_model)}.",
                BaseModelWarning,
            )
        return _lmfit_model

    def __repr__(self):

        _choice = "no" if not self.has_substrate else "yes"
        _txt = f"{self.model_name}, substrate ({_choice}): "
        if hasattr(self, "lmfit_model"):
            _txt += "\n\t" + repr(self.lmfit_model)
        else:
            _txt += "empty model"
        return _txt

    # def _0equalize_from_model_name(self,_name):
    #     pass
    # def _0equalize_from_incl_substrate(self,_choice):
    #     pass
    # def _0equalize_name_choice(self, _name, _choice):
    #     _change = False
    #     if _choice != None and _name == None and hasattr(self, '_model_name'):
    #         # change model name when choice is set
    #         _name = self._add_or_rem_substrate_to_model_name(_choice, self._model_name)
    #         if _name != self._model_name:
    #             _change = True
    #             self._model_name = _name
    #     if _name != None and _choice == None and hasattr(self, '_include_substrate'):
    #         # change include subtrate choice when model name is set
    #         _name_contains = self._name_contains_substrate(_name)
    #         if _name_contains != self._include_substrate:
    #             _change = True
    #             self._include_substrate = _name_contains
    #         if hasattr(self,'_model_name'):
    #             _change = bool(_name != self._model_name)
    #     # if _name != None and hasattr(self, '_model_name')
    #     print(f'Change {_change}, name {_name}, choice {_choice}')
    #     if _change  and _name:
    #         self.lmfit_model = self.model_constructor_from_model_name(_name)
    #     elif not _choice and _contains:
    #         _substr_name = self.substrate_peak_name.split(self._SUFFIX)[0]
    #         warn(f'\n{self.__class__.__name__} include substrate is set to {_choice} so "{_substr_name}" is removed from {_name}.\n',BaseModelWarning)
    #         _name = '+'.join(i for i in _name.split('+') if i not in _substr_name)  # remove substr name

    # def _0add_or_rem_substrate_to_model_name(self, _choice, _name):
    #     _substr_name = self.substrate_peak_name.split(self._SUFFIX)[0]
    #     _contains = self._name_contains_substrate(_name)
    #     if _choice and not _contains:
    #         _name = _name+f'+{_substr_name}'  # add substr name
    #     elif not _choice and _contains:
    #         warn(f'\n{self.__class__.__name__} include substrate is set to {_choice} so "{_substr_name}" is removed from {_name}.\n',BaseModelWarning)
    #         _name = '+'.join(i for i in _name.split('+') if i not in _substr_name)  # remove substr name
    #     return _name



import logging
from collections import OrderedDict
from warnings import warn


logger = logging.getLogger(__name__)

logging.captureWarnings(True)  # sends these warning to the logger


class FieldsTrackerWarning(UserWarning):
    pass


class FieldsTracker:
    """
    Keeps check of the fields from multiple sources,
    allows to store values in dict
    yields results a single results from several sources for each field
    status is True when all fields in results have at least one value
    """

    def __init__(self, fields: list = [], sources: tuple = [], **kwargs):
        self.fields = fields
        self.sources = sources
        self._register_template = self.make_register(sources, fields)
        self.set_sources_attr()
        self._results = {}

    def make_register(self, sources, fields):
        _reg = {source: {field: None for field in fields} for source in sources}
        return _reg

    def set_sources_attr(self):
        for source in self.sources:
            setattr(self, f"{source}", self._register_template[source])

    @property
    def register(self):
        _reg = {source: getattr(self, source) for source in self.sources}
        return _reg

    @property
    def status(self):
        _st = False
        if set(self.results) == set(self.fields):
            _st = True
        return _st

    @property
    def results(self):
        return self._results

    def _set_results(self):
        _results = self.get_values_from_all_fields()
        self._results = _results

    @property
    def missing(self):
        results = self.results
        _missing = set(self.fields) - set(results.keys())
        return _missing

    def get_values_from_all_fields(self):
        _result_values = {}
        for field in self.fields:
            _fvaldict_sources = self.get_field_value_from_sources(field)
            if _fvaldict_sources:
                _src = {"source": i for i in _fvaldict_sources.keys()}
                _value = {"value": i for i in _fvaldict_sources.values()}
                _nice_result = {**_src, **_value}
                _result_values.update({field: _nice_result})
        return _result_values

    def get_field_value_from_sources(self, field):
        _fsvals = OrderedDict({})
        _result = {}
        for source in self.sources:
            _src = getattr(self, source)
            _fval = _src.get(field, None)
            if _fval:
                _fsvals.update({source: _fval})
        # if _fsvals:
        #     if not dict in map(type, _fsvals.values()):
        #         breakpoint()
        #         _setvals = set(_fsvals.values())
        #     else:
        _setvals = _fsvals.values()
        _setsources = set(_fsvals.keys())
        _lstsources = list(_setsources)
        if len(_setvals) == 1:
            _fval = list(_setvals)[0]
            if len(_setsources) == 1:
                _src = _lstsources[0]
            elif len(_setsources) > 1:
                _src = list(_fsvals.keys())[0]
                warn(
                    f"Field {field} has multiple sources {_setsources}, one value ",
                    FieldsTrackerWarning,
                )
            _result = {_src: _fval}
        elif len(_setvals) > 1:
            # breakpoint()
            _firstval = list(_fsvals.items())[0]
            warn(
                f"Field {field} has multiple sources {_setsources}, different values follow order of sources ",
                FieldsTrackerWarning,
            )
            _result = {_firstval[0]: _firstval[1]}
        return _result

    def multi_store(self, source: str, **kwargs):
        # _fields_dict = {k: val for k, val in _dict.items() if k in self.fields}
        _fields_kwargs = {k: val for k, val in kwargs.items() if k in self.fields}
        # _input_dict = {**_fields_kwargs, **_fields_dict}
        if _fields_kwargs:
            for field, val in _fields_kwargs.items():
                self.store(source, field, val)
            self._set_results()

    def store(self, source, field, val):
        """store one value: source, field, val"""
        if source in self.sources and field in self.fields and val:
            _src = getattr(self, source)
            _fval = _src.get(field, None)
            if not _fval:
                _src[field] = val
            elif _fval == val:
                warn(
                    f"Redefinition of {field} in {source} ignored",
                    FieldsTrackerWarning,
                )
            elif _fval != val:
                _src[field] = val
                warn(
                    f"Overwriting of {field} in {source} with new value! {_fval} is not {val}",
                    FieldsTrackerWarning,
                )
            else:
                warn(f"Store {source} {val} unexpected", FieldsTrackerWarning)

            setattr(self, source, _src)
            self._set_results()
        else:
            warn(
                f"Store in {source} at {field} not in {self.sources} or not in {self.fields} or not {val}, ignored.",
                FieldsTrackerWarning,
            )
            pass  # store values not recognized



import copy
import logging
import unittest

import pytest
from lmfit import Model

# from raman_fitting.deconvolution_models import first_order_peaks
import raman_fitting
from raman_fitting.deconvolution_models.default_peaks.base_peak import (
    BasePeak,
    BasePeakWarning,
    LMfitModelConstructorMethods,
)

logger = logging.getLogger(__name__)
logging.captureWarnings(True)  # sends these warning to the logger


#%%


def _error_message_contains(excinfo, testmsg: str, verbose: bool = False):
    _fltr_str = [
        i if not i in ["(", ")"] else " "
        for i in str(excinfo.value)
        if i.isalnum() or i in (",", ".", " ", "_", "(", ")")
    ]
    _cl_str = "".join(map(str, _fltr_str))
    _cl_str_split = _cl_str.split(" ")
    _test = all(i in _cl_str_split for i in testmsg.split(" "))

    if not _test:
        _test = any(i in _cl_str_split for i in testmsg.split(" "))

    if not _test or verbose:
        print(list(((i, i in _cl_str_split) for i in testmsg.split(" "))))
        print(_cl_str_split)
    return _test


#%%


class TestBasePeak(unittest.TestCase):

    #%% TESTING

    def test_BasePeak_attributes(self):
        self.assertTrue(BasePeak.__doc__)
        self.assertTrue(BasePeak._fields)
        self.assertTrue(BasePeak._sources)
        self.assertTrue(BasePeak.PEAK_TYPE_OPTIONS)

    def test_empty_base_class_raises(self):
        #%%

        class EmptyTestChild(metaclass=BasePeak):
            pass

        eb = EmptyTestChild()

        with pytest.raises(ValueError) as excinfo:
            eb.peak_type = "emptytest"
        assert _error_message_contains(excinfo, "value emptytest for peak_type not in")

        with pytest.raises(ValueError) as excinfo:
            eb.peak_name = 10 * "emptytest"
        assert _error_message_contains(excinfo, "value for peak_name is too long 90")

        self.assertFalse(eb.peak_model)

        #%%

    def test_empty_base_class_with_kwargs_raises(self):
        #%%

        class EmptyTestChild(
            metaclass=BasePeak,
            testkwarg2=2,
            testkwarg3=3,
            peak_type="Voigt",
            verbose=True,
        ):
            pass

        eb = EmptyTestChild()
        #%%

        self.assertEqual(eb.peak_type, "Voigt")
        self.assertEqual(eb.testkwarg2, 2)

        with pytest.raises(ValueError) as excinfo:
            eb.peak_type = "emptytest"
        self.assertTrue(
            _error_message_contains(excinfo, "value emptytest for peak_type not in")
        )

        with pytest.raises(ValueError) as excinfo:
            eb.peak_name = 10 * "emptytest"
        self.assertTrue(
            _error_message_contains(excinfo, "value for peak_name is too long 90")
        )

        self.assertFalse(eb.peak_model)
        # with pytest.raises(AttributeError) as excinfo:
        #     eb.peak_model
        # self.assertTrue(
        #     _error_message_contains(
        #         excinfo, "type object empty no attribute _peak_model"
        #     )
        # )

        with pytest.raises(ValueError) as excinfo:
            eb.peak_type = "VoigtLorentzian"
        self.assertTrue(
            _error_message_contains(
                excinfo,
                ''''Multiple options ['Lorentzian', 'Voigt'] for misspelled value "VoigtLorentzian"''',
            )
        )

    def test_empty_base_class_with_false_input(self):
        #%%

        class EmptyTestChild(metaclass=BasePeak, peak_type="FalsePeak"):
            pass

        with pytest.raises(ValueError) as excinfo:
            eb = EmptyTestChild()
        self.assertTrue(
            _error_message_contains(excinfo, "value emptytest for peak_type not in")
        )

    #%%
    def test_base_class_good_with_init_extra_tests(self):
        #%%
        class TestD1peak(metaclass=BasePeak, debug=True):
            """
            here is docstring of TestD1peak,
            small spelling error on peak_type
            """

            def __init__(self, *args, **kwargs):
                print(f"called __init__ {self} TestD1peak, with {args}, {kwargs}")
                self._peak_type = "Voigt"
                self.peak_name = "D1D1"
                self.input_param_settings = {
                    "center": {"value": 2650, "min": 2600, "max": 2750},
                    "sigma": {"value": 60, "min": 1, "max": 200},
                    "amplitude": {"value": 14, "min": 1e-03, "max": 100},
                }

        td1 = TestD1peak()
        peakmod = "<lmfit.Model: Model(voigt, prefix='D1D1_')>"
        self.assertEqual(str(td1.peak_model), peakmod)
        _class_str = f"TestD1peak, {peakmod}, center : 2600 < 2650 > 2750"
        self.assertIn(_class_str, str(td1))
        td1.peak_name = "R2D2"
        self.assertEqual(td1.peak_model.prefix, "R2D2_")

        # _def_param = td1.param_hints_constructor({})
        # _def_key = list(BasePeak.default_settings.keys())[0]
        # _def_param[_def_key].value == BasePeak.default_settings[_def_key]["value"]

        # _def_param = td1.param_hints_constructor(
        #     td1.fco.register["init"]["param_hints"]
        # )
        # self.assertEqual(_def_param["amplitude"].value, 14)

        # with pytest.raises(TypeError) as excinfo:
        #     _def_param = td1.param_hints_constructor("fail")
        # self.assertTrue(
        #     _error_message_contains(
        #         excinfo,
        #         "input_param_hints should be of type dictionary not <class 'str'>",
        #     )
        # )

        # _err_hints = copy.copy(td1.fco.register["init"]["param_hints"])
        # _err_hints["center"] = (1, 2, 3, 4)

        # with pytest.raises(ValueError) as excinfo:
        #     _def_param = td1.param_hints_constructor(_err_hints)
        # self.assertTrue(
        #     _error_message_contains(
        #         excinfo, " Unable to create a Parameter from center and (1, 2, 3, 4):"
        #     )
        # )

    #%%
    def test_base_class_good_with_init(self):
        class TestD1peak(metaclass=BasePeak, debug=True):
            """
            test_base_class_good_with_init
            but with spelling error in peak_type
            """

            def __init__(self, *args, **kwargs):
                print(f"called __init__ {self} TestD1peak, with {args}, {kwargs}")
                self._peak_type = "cartVoigt"
                self.peak_name = "D1D1"
                self.input_param_settings = {
                    "center": {"value": 2650, "min": 2600, "max": 2750},
                    "sigma": {"value": 60, "min": 1, "max": 200},
                    "amplitude": {"value": 14, "min": 1e-03, "max": 100},
                }

        td1 = TestD1peak()
        _class_str = "TestD1peak, <lmfit.Model: Model(voigt, prefix='D1D1_')>, center : 2600 < 2650 > 2750"
        self.assertIn(_class_str, str(td1))
        # print(td1)

    #%%

    def test_base_class_good_with_init_added_method(self):
        #%%
        class TestD1peakmeta(metaclass=BasePeak, a=2):
            """
            here is docstring of TestD1peak
            """

            def __init__(self, *args, **kwargs):
                # super().__init__(self)
                # print(f'__subclass __init__ self "{self}" TestD1peakmeta, with {args}, {kwargs}')
                self.peak_type = "Lorentzian"
                self.peak_name = "D1D1"
                self.input_param_settings = {
                    "center": {"value": 2650, "min": 2600, "max": 2750},
                    "sigma": {"value": 60, "min": 1, "max": 200},
                    "amplitude": {"value": 14, "min": 1e-03, "max": 100},
                }
                self._meta_added_method(kwargs)

            def _meta_added_method(self, *args, **kwargs):
                """added method"""
                print(f"{self.__dict__}")
                self._added_method_arg = kwargs
                print(f"added method called {kwargs}")
                # return arg

        td1m = TestD1peakmeta(add=33)
        _teststr = "TestD1peakmeta, <lmfit.Model: Model(lorentzian, prefix='D1D1_')>, center : 2600 < 2650 > 2750"
        self.assertIn(_teststr, str(td1m))
        # assert str(td1m) == _teststr

        #%%

    def test_base_class_good_with_attributes_and_init(self):
        #%%
        class NewChildClassAttr(metaclass=BasePeak):
            """New child class for easier definition"""

            _test = "testkwarg"

            param_hints = {
                "center": {"value": 2435, "min": 2400, "max": 2550},
                "sigma": {"value": 30, "min": 1, "max": 200},
                "amplitude": {"value": 2, "min": 1e-03, "max": 100},
            }
            peak_type = "Voigt"  #'Voigt'
            peak_name = "R2D2"

            def __init__(self, **kwargs):
                # print(f'__init child empty pass {self}')
                pass
                # super().__init__()

        nca = NewChildClassAttr()
        _center_value = nca.peak_model.param_hints["center"]["value"]
        assert _center_value == 2435
        # print('Instance child:', nca)
        #%%

    def test_base_class_good_with_attributes_no_init(self):
        class NewChildClassAttrNoInit(metaclass=BasePeak):
            """New child class for easier definition"""

            _test = "testkwarg"

            param_hints = {
                "center": {"value": 2435, "min": 2400, "max": 2550},
                "sigma": {"value": 30, "min": 1, "max": 200},
                "amplitude": {"value": 2, "min": 1e-03, "max": 100},
            }
            peak_type = "Voigt"  #'Voigt'
            peak_name = "R2D2"
            # def __init__(self,**kwargs):
            # print(f'__init child empty pass {self}')
            # pass
            # super().__init__()

        ncni = NewChildClassAttrNoInit()
        _center_value = ncni.peak_model.param_hints["center"]["value"]
        assert _center_value == 2435

        #%%

    def test_base_class_good_with_attributes_init_collision_values(self):
        #%%
        class NewChildClassInit(metaclass=BasePeak, peak_type="Gaussian"):
            """New child class for easier definition"""

            _test = "testkwarg"
            peak_type = "Lorentzian"
            param_hints = {
                "center": {"value": 2435, "min": 2400, "max": 2550},
                "sigma": {"value": 30, "min": 1, "max": 200},
                "amplitude": {"value": 2, "min": 1e-03, "max": 100},
            }

            def __init__(self, **kwargs):
                self.peak_type = "Voigt"
                self.peak_name = "R2D2"
                self.param_hints = {"center": {"value": 20}}

                # self.param_hints = {
                #                     'center':
                #                         {'value': 2435,'min': 2400, 'max': 2550},
                #                     'sigma':
                #                         {'value': 30,'min' : 1, 'max': 200},
                #                     'amplitude' :
                #                         {'value': 2,'min' : 1E-03, 'max': 100}
                #                         }
                # super().__init__(self)

        nci = NewChildClassInit()
        self.assertEqual(nci.peak_type, "Gaussian")
        self.assertEqual(nci.peak_model.param_hints["center"]["value"], 2435)
        # print(nci)
        # print(self.__dict__)

    def test_base_with_only_keyword_args(self):
        new = BasePeak("newPeak", **{"noname": 2, "debug": False, "peak_type": "Voigt"})
        new.param_hints = {"center": {"value": 200}}
        newinst = new()
        _newinst_str = "newPeak, <lmfit.Model: Model(voigt, prefix='newPeak_')>, center : -inf < 200 > inf"
        self.assertEqual(str(newinst), _newinst_str)


class TestLMfitModelConstructorMethods(unittest.TestCase):

    LMfit = LMfitModelConstructorMethods

    def test_make_model_from_peak_type_and_name(self):

        model = self.LMfit.make_model_from_peak_type_and_name(
            peak_type="Voigt", peak_name="lmfitpeak"
        )
        self.assertTrue(isinstance(model, Model))
        self.assertEqual(model.prefix, "lmfitpeak")

        with pytest.raises(NotImplementedError) as excinfo:
            model = self.LMfit.make_model_from_peak_type_and_name(peak_type="FalsePeak")
        self.assertTrue(
            _error_message_contains(
                excinfo,
                " This peak type or model 'FalsePeak' has not been implemented.",
            )
        )


# self = TestLMfitModelConstructorMethods()


#%%
if __name__ == "__main__":
    unittest.main()
    self = TestBasePeak()



""" Class for reading in files, can be extended for other than txt formats"""

from pathlib import Path

import numpy as np


class FileReader:
    def __init__(self, file_path=Path()):
        self._file_path = file_path
        self.read_in()

    def read_in(self):
        ramanshift, intensity_raw = np.array([]), np.array([])
        i = 0
        while not ramanshift.any():
            try:
                ramanshift, intensity_raw = np.loadtxt(
                    self._filepath, usecols=(0, 1), unpack=True, skiprows=i
                )
                print(self._filepath, len(ramanshift), len(intensity_raw))
                self._skiprows = i
                self._read_succes = True
            except ValueError:
                i += 1

        self.ramanshift = ramanshift
        self.intensity_raw = intensity_raw



# flake8: noqa

import unittest

import pytest
from lmfit import Model

import raman_fitting
from raman_fitting.deconvolution_models.peak_validation import (
    NotFoundAnyModelsWarning,
    PeakModelValidator,
)


class TestPeakModelValidator(unittest.TestCase):
    def setUp(self):
        self.pmv = PeakModelValidator()

    def test_pmv_valid_models(self):
        self.assertTrue(self.pmv.valid_models)

    def test_pmv_set_debug(self):
        self.assertFalse(self.pmv.debug)
        self.assertTrue(self.pmv._set_debug(**{"debug": True}))

    def test_get_subclasses_from_base(self):
        with self.assertWarns(NotFoundAnyModelsWarning):
            self.pmv.get_subclasses_from_base("")

        with self.assertWarns(NotFoundAnyModelsWarning):
            self.pmv.get_subclasses_from_base(str)

    def test_validation_inspect_models(self):

        _valid = self.pmv.validation_inspect_models([str])
        self.assertTrue(_valid)
        self.assertFalse(_valid[1][0].valid)
        self.assertIn("has no attr", _valid[1][0].message)

        _valid = self.pmv.validation_inspect_models([Model])
        self.assertTrue(_valid)
        self.assertFalse(_valid[1][0].valid)
        self.assertIn("Unable to initialize model", _valid[1][0].message)

    def test_get_cmap_list(self):

        _cmap = self.pmv.get_cmap_list([], cmap_options=())
        self.assertEqual(_cmap, [])
        _cmap = self.pmv.get_cmap_list(
            [1] * 50, cmap_options=(), fallback_color=self.pmv.fallback_color
        )
        self.assertEqual(_cmap, [self.pmv.fallback_color] * 50)
        _cmap = self.pmv.get_cmap_list([1] * 5)
        self.assertEqual(len(_cmap), 5)

    def test___getattr__(self):

        with self.assertRaises(AttributeError):
            self.pmv.fake_attr

    def test___iter__(self):
        _iter = [i for i in self.pmv]
        self.assertIsInstance(_iter, list)

    def test_if_lmfit_models(self):

        if self.pmv.lmfit_models:
            _getdict = self.pmv.get_dict()
            self.assertIsInstance(_getdict, dict)

            _getdict = self.pmv.get_model_dict(self.pmv.lmfit_models)
            self.assertIsInstance(_getdict, dict)


def _debugging():

    self = TestPeakModelValidator()
    peaks = PeakModelValidator()
    self.pmv = peaks
    # all([isinstance(i.peak_model, Model) for i in peaks.lmfit_models])
    # all([isinstance(i.peak_model, Model) for i in peaks.get_dict().values()])


if __name__ == "__main__":
    unittest.main()



# flake8: noqa
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from collections import namedtuple

import numpy as np
import pandas as pd

# from raman_fitting.processing.slicer import SpectraInfo
# from raman_fitting.processing.cleaner import SpectrumCleaner


def _testing():
    spec_raw = sample_spectra[1]
    spec = spec_raw
    windowname = "1st_order"

    spec = norm_spec



# import datetime
import hashlib
import logging
from pathlib import Path

from typing import Dict, List

# from .. import __package_name__

from .filedata_parser import SpectrumReader
from .filename_parser_helpers import filestem_to_sid_and_pos, sID_to_sgrpID, get_fstats

logger = logging.getLogger(__name__)

#%%
# _extra_sID_name = 'Si-ref'
index_primary_key = "rfID"
index_file_primary_keys = {f"{index_primary_key}": "string"}
index_file_path_keys = {"FileStem": "string", "FilePath": "Path"}
index_file_sample_keys = {
    "SampleID": "string",
    "SamplePos": "int64",
    "SampleGroup": "string",
}
index_file_stat_keys = {
    "FileCreationDate": "datetime64",
    "FileCreation": "float",
    "FileModDate": "datetime64",
    "FileMod": "float",
    "FileSize": "int64",
}
index_file_read_text_keys = {"FileHash": "string", "FileText": "string"}

index_dtypes_collection = {
    **index_file_path_keys,
    **index_file_sample_keys,
    **index_file_stat_keys,
    **index_file_read_text_keys,
}

# Extra name to sID mapper, if keys is in filename
_extra_sID_name_mapper = {
    "David": "DW",
    "stephen": "SP",
    "Alish": "AS",
    "Aish": "AS",
}

# Extra name to sID mapper, if key is in filepath parts
_extra_sgrpID_name_mapper = {"Raman Data for fitting David": "SH"}


def _extra_overwrite_sID_from_mapper(
    sID: str, mapper: dict = _extra_sID_name_mapper
) -> str:
    """Takes an sID and potentially overwrites from a mapper dict"""
    _sID_map = mapper.get(sID, None)
    if _sID_map:
        sID = _sID_map
    return sID


def _extra_overwrite_sgrpID_from_parts(
    parts: List[str], sgrpID: str, mapper: dict = _extra_sgrpID_name_mapper
) -> str:
    for k, val in _extra_sgrpID_name_mapper.items():
        if k in parts:
            sgrpID = val
    return sgrpID


class PathParser(Path):
    """
    This class parses the filepath of a file to a parse_result (dict),
    from which the main raman file index will be built up.

    """

    _flavour = type(Path())._flavour

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._qcnm = self.__class__.__qualname__
        self.stats_ = None
        self.data = None
        self.parse_result = self.collect_parse_results(**kwargs)

    @staticmethod
    def get_rfID_from_path(path: Path) -> str:
        """
        Makes the ID from a filepath

        Parameters
        ----------
        path : Path
            DESCRIPTION.

        Returns
        -------
        str: which contains hash(parent+suffix)_stem of path

        """

        _parent_suffix_hash = hashlib.sha512(
            (str(path.parent) + path.suffix).encode("utf-8")
        ).hexdigest()
        _filestem = path.stem
        fnID = _parent_suffix_hash + "_" + _filestem
        return fnID

    def collect_parse_results(
        self, read_data=False, store_data=False, **kwargs
    ) -> Dict:
        """performs all the steps for parsing the filepath"""
        parse_res_collect = {}

        if self.exists():
            if self.is_file():
                self.stats_ = self.stat()

                _fnID = self.make_dict_from_keys(
                    index_file_primary_keys, (self.get_rfID_from_path(self),)
                )
                _filepath = self.make_dict_from_keys(
                    index_file_path_keys, (self.stem, self)
                )
                _sample = self.parse_sample_with_checks()
                _filestats = self.parse_filestats(self.stats_)
                if read_data == True:
                    try:
                        self.data = SpectrumReader(self)
                    except Exception as exc:
                        logger.warning(
                            f"{self._qcnm} {self} SpectrumReader failed.\n{exc}"
                        )

                parse_res_collect = {**_fnID, **_filepath, **_sample, **_filestats}
            else:
                logger.warning(f"{self._qcnm} {self} is not a file => skipped")
        else:
            logger.warning(f"{self._qcnm} {self} does not exist => skipped")
        return parse_res_collect

    def parse_sample_with_checks(self):
        """parse the sID, position and sgrpID from stem"""

        _parse_res = filestem_to_sid_and_pos(self.stem)

        if len(_parse_res) == 2:
            sID, position = _parse_res

            try:
                sID = _extra_overwrite_sID_from_mapper(sID)
            except Exception as exc:
                logger.info(
                    f"{self._qcnm} {self} _extra_overwrite_sID_from_mapper failed => skipped.\n{exc}"
                )

            sgrpID = sID_to_sgrpID(sID)

            try:
                sgrpID = _extra_overwrite_sgrpID_from_parts(self.parts, sgrpID)
            except Exception as exc:
                logger.info(
                    f"{self._qcnm} {self} _extra_overwrite_sgrpID_from_parts failed => skipped.\n{exc}"
                )

            _parse_res = sID, position, sgrpID
        else:
            logger.warning(
                f"{self._qcnm} {self} failed to parse filename to sID and position."
            )
        return self.make_dict_from_keys(index_file_sample_keys, _parse_res)

    def parse_filestats(self, fstat) -> Dict:
        """get status metadata from a file"""

        filestats = get_fstats(fstat)
        return self.make_dict_from_keys(index_file_stat_keys, filestats)

    def make_dict_from_keys(self, _keys_attr: Dict, _result: tuple) -> Dict:
        """returns dict from tuples of keys and results"""
        if not isinstance(_result, tuple):
            logger.warning(
                f"{self._qcnm} input value is not a tuple, {_result}. Try to cast into tuple"
            )
            _result = (_result,)

        _keys = _keys_attr.keys()

        if not len(_result) == len(_keys) and not isinstance(_keys, str):
            # if len not matches make stand in numbered keys
            _keys = [f"{_keys_attr}_{n}" for n, i in enumerate(_result)]
        return dict(zip(_keys, _result))



""" Indexer for raman data files """
import hashlib
from typing import List

# get_directory_paths_for_run_mode
# from .index_selection import index_selection
import logging
import sys
from pathlib import Path

import pandas as pd

from raman_fitting.config.filepath_helper import get_directory_paths_for_run_mode

# parse_filepath_to_sid_and_pos
from raman_fitting.indexing.filename_parser import index_dtypes_collection
from raman_fitting.indexing.filename_parser_collector import make_collection

# from raman_fitting.utils._dev_sqlite_db import df_to_db_sqlalchemy

# from .. import __package_name__


logger = logging.getLogger(__name__)
logger.propagate = False

__all__ = ["MakeRamanFilesIndex"]

#%%


class MakeRamanFilesIndex:
    """

    Finds the RAMAN files in the data folder from config and creates an overview, on the attribute .index
    finds a list of files,

    """

    # index_file_sample_cols = {'FileStem': 'string',
    #                           'SampleID': 'string',
    #                           'SamplePos': 'int64',
    #                           'SampleGroup': 'string',
    #                           'FilePath': 'string')
    # index_file_stat_cols = ('FileCreationDate' , 'FileCreation','FileModDate', 'FileMod', 'FileHash')
    # INDEX_FILE_NAME = 'index.csv'
    debug = False

    table_name = "ramanfiles"

    # RESULTS_DIR = config.RESULTS_DIR,
    #              DATASET_DIR = config.DATASET_DIR,
    #              INDEX_FILE = config.INDEX_FILE,
    def __init__(
        self, force_reload=True, run_mode="normal", dataset_dirs=None, **kwargs
    ):

        self._cqnm = self.__class__.__qualname__

        self._kwargs = kwargs
        self.force_reload = force_reload
        self.run_mode = run_mode

        if not dataset_dirs:
            dataset_dirs = get_directory_paths_for_run_mode(run_mode=self.run_mode)

        self.dataset_dirs = dataset_dirs
        for k, val in self.dataset_dirs.items():
            if isinstance(val, Path):
                setattr(self, k, val)
                # if val.is_dir() or val.is_file():

        self.raman_files = self.find_files(data_dir=self.DATASET_DIR)
        self.index = pd.DataFrame()
        self._error_parse_filenames = []
        if "normal" in run_mode and not self.debug and not self.force_reload:
            self.index = self.load_index()

        else:
            self.index = self.reload_index()

        self.index_selection = self.index_selection(self.index, **self._kwargs)

    @staticmethod
    def find_files(data_dir: Path = Path()) -> List:
        """
        Creates a list of all raman type files found in the DATASET_DIR which are used in the creation of the index.
        """

        if not isinstance(data_dir, Path):
            logger.warning(f"find_files warning: arg is not Path.")
            return []

        raman_files_raw = []
        if data_dir.exists():
            RFs = data_dir.rglob("*txt")
            if RFs:
                raman_files_raw = [
                    i
                    for i in RFs
                    if not "fail" in i.stem and not "Labjournal" in str(i)
                ]
                logger.info(
                    f"find_files {len(raman_files_raw)} files were found in the chosen data dir:\n\t{data_dir}"
                )
            else:
                logger.warning(
                    f"find_files warning: the chose data file dir was empty.\n{data_dir}\mPlease choose another directory which contains your data files."
                )
        else:
            logger.warning(
                f"find_files warning: the chosen data file dir does not exists.\n{data_dir}\nPlease choose an existing directory which contains your data files."
            )

        return raman_files_raw

    def make_index(self):
        """loops over the files and scrapes the index data from each file"""
        raman_files = self.raman_files
        pp_collection = make_collection(raman_files, **self._kwargs)

        index = pd.DataFrame([i.parse_result for i in pp_collection])
        index = self._extra_assign_destdir_and_set_paths(index)
        logger.info(
            f"{self._cqnm} successfully made index {len(index)} from {len(raman_files)} files"
        )
        if self._error_parse_filenames:
            logger.info(
                f"{self._cqnm} errors for filename parser {len(self._error_parse_filenames)} from {len(raman_files)} files"
            )
        return index

    def _extra_assign_destdir_and_set_paths(self, index: pd.DataFrame):
        """assign the DestDir column to index and sets column values as object type"""

        if hasattr(index, "SampleGroup"):
            index = index.assign(
                **{
                    "DestDir": [
                        self.RESULTS_DIR.joinpath(sGrp)
                        for sGrp in index.SampleGroup.to_numpy()
                    ]
                }
            )
        _path_dtypes_map = {
            k: val for k, val in index_dtypes_collection.items() if "Path" in val
        }
        for k, val in _path_dtypes_map.items():
            if hasattr(index, k):
                if "Path" in val:
                    index[k] = [Path(i) for i in index[k].to_numpy()]
        return index

    def export_index(self, index):
        """saves the index to a defined Index file"""
        if not index.empty:
            if not self.INDEX_FILE.parent.exists():
                logger.info(
                    f"{self._cqnm} created parent dir: {self.INDEX_FILE.parent}"
                )
                self.INDEX_FILE.parent.mkdir(exist_ok=True, parents=True)

            index.to_csv(self.INDEX_FILE)

            _dtypes = index.dtypes.to_frame("dtypes")
            _dtypes.to_csv(self._dtypes_filepath())

            # self.save_merge_to_db(DB_filepath, index, self.table_name)

            logger.info(
                f"{self._cqnm} Succesfully Exported Raman Index file to:\n\t{self.INDEX_FILE}\nwith len({len(index)})."
            )
        else:
            logger.info(f"{self._cqnm} Empty index not exported")

    def load_index(self):
        """loads the index from from defined Index file"""
        if self.INDEX_FILE.exists():
            try:

                _dtypes = pd.read_csv(self._dtypes_filepath(), index_col=[0]).to_dict()[
                    "dtypes"
                ]

                _dtypes_datetime = {
                    k: val
                    for k, val in _dtypes.items()
                    if "datetime" in val or k.endswith("Date")
                }

                _dtypes_no_datetime = {
                    k: val
                    for k, val in _dtypes.items()
                    if k not in _dtypes_datetime.keys()
                }

                index = pd.read_csv(
                    self.INDEX_FILE,
                    index_col=[0],
                    dtype=_dtypes_no_datetime,
                    parse_dates=list(_dtypes_datetime.keys()),
                )
                index = self._extra_assign_destdir_and_set_paths(index)

                logger.info(
                    f"Succesfully imported Raman Index file from {self.INDEX_FILE}, with len({len(index)})"
                )
                if not len(self.index) == (
                    len(self.raman_files) + len(self._error_parse_filenames)
                ):
                    logger.error(
                        f"""'Error in load_index from {self.INDEX_FILE},
                                 \nlength of loaded index not same as number of raman files
                                 \n starting reload index ... """
                    )
                    self.index = self.reload_index()

            except Exception as e:
                logger.error(
                    f"Error in load_index from {self.INDEX_FILE},\n{e}\n starting reload index ... "
                )
                index = self.reload_index()
        else:
            logger.error(
                f"Error in load_index: {self.INDEX_FILE} does not exists, starting reload index ... "
            )
            index = self.reload_index()
        return index

    def reload_index(self):
        """restarts the index creation from scratch and export."""
        logger.info(f"{self._cqnm} starting reload index.")
        index = pd.DataFrame()

        try:
            logger.info(f"{self._cqnm} making index.")

            try:
                index = self.make_index()
            except Exception as e:
                logger.error(f"{self._cqnm} make index error:\n\t{e}")

            try:
                self.export_index(index)
            except Exception as e:
                logger.error(f"{self._cqnm} export after make index error:\n\t{e}")

        except Exception as e:
            logger.error(f"{self._cqnm} reload index error:\n\t{e}")

        return index

    def index_selection(
        self, index=pd.DataFrame(), default_selection: str = "", **kwargs
    ):
        """
        Special selector on the index DataFrame

        Parameters
        -------

        index
            pd.DataFrame containing the index of files
            should contains columns that are given in index_file_sample_cols and index_file_stat_cols
        default_selection str
            all or '' for empty default
        kwargs
            checks for keywords suchs as samplegroups, sampleIDs, extra
            meant for cli commands

        Returns
        -------
        index_selection
            pd.DataFrame with a selection from the given input parameter index
            default returns empty DataFrame

        """

        _kws = kwargs
        _keys = _kws.keys()

        default_selection = _kws.get("default_selection", default_selection)
        if not "normal" in _kws.get("run_mode", default_selection):
            default_selection = "all"
        index_selection = pd.DataFrame()
        logger.info(
            f"{self._cqnm} starting index selection from index({len(index)}) with:\n default selection: {default_selection}\n and {kwargs}"
        )

        if not index.empty:

            if default_selection:
                if default_selection == "all":
                    index_selection = index.copy()

            if "samplegroups" in _keys:
                if _kws["samplegroups"]:
                    index_selection = index.loc[
                        index.SampleGroup.str.contains("|".join(_kws["samplegroups"]))
                    ]
            if "sampleIDs" in _keys:
                index_selection = index.loc[
                    index.SampleID.str.contains("|".join(_kws["sampleIDs"]))
                ]

            if "extra" in _keys:
                runq = _kws.get("run")
                if "recent" in runq:
                    grp = index.sort_values(
                        "FileCreationDate", ascending=False
                    ).FileCreationDate.unique()[0]

                    index_selection = index.loc[index.FileCreationDate == grp]
                    index_selection = index_selection.assign(
                        **{
                            "DestDir": [
                                Path(i).joinpath(grp.strftime("%Y-%m-%d"))
                                for i in index_selection.DestDir.values
                            ]
                        }
                    )

            if "make_examples" in self.run_mode:
                index_selection = index.loc[~index.SampleID.str.startswith("Si")]

            logger.debug(
                f"{self._cqnm} finished index selection from index({len(index)}) with:\n {default_selection}\n and {kwargs}\n selection len({len(index_selection )})"
            )
        else:
            logger.warning(f"{self._cqnm} index selection index arg empty")

        if index_selection.empty:
            logger.warning(f"{self._cqnm} index selection empty. exiting")
            sys.exit()

        return index_selection

    def _dtypes_filepath(self):
        _dtypes_filepath = self.INDEX_FILE.with_name(
            self.INDEX_FILE.stem + "_dtypes" + self.INDEX_FILE.suffix
        )
        return _dtypes_filepath

    def __repr__(self):
        return f"{self._cqnm} with index ({len(self.index)})"

    def __len__(self):
        return len(self.index)


def main():
    """test run for indexer"""
    RamanIndex = None
    try:
        RamanIndex = MakeRamanFilesIndex(read_data=True, run_mode="make_examples")

    except Exception as e:
        logger.error(f"Raman Index error: {e}")
    return RamanIndex


if __name__ == "__main__":
    # print(getattr(__file__, 'testdt'))
    RamanIndex = main()



import datetime
import unittest
from importlib import resources
from pathlib import Path

# from raman_fitting.deconvolution_models import first_order_peaks
import pandas as pd
import pytest

import raman_fitting
from raman_fitting.datafiles import example_files
from raman_fitting.indexing.indexer import MakeRamanFilesIndex


class TestIndexer(unittest.TestCase):
    def setUp(self):

        _example_path = Path(example_files.__path__[0])
        _example_files_contents = list(Path(_example_path).rglob("*txt"))

        self._example_files = [i for i in _example_files_contents]

        self.RamanIndex = MakeRamanFilesIndex(run_mode="make_examples")

    def test_MakeRamanFilesIndex_make_examples(self):

        self.assertEqual(len(self.RamanIndex), len(self._example_files))

    def test_load_index(self):
        _loaded_index = self.RamanIndex.load_index()
        self.assertTrue(isinstance(_loaded_index, pd.DataFrame))

        for col in _loaded_index.columns:
            _setload = set(_loaded_index[col].values)

            _setindex = set(self.RamanIndex.index[col].values)
            if all(isinstance(i, datetime.date) for i in list(_setindex)):
                # Convert pandas, np.datetime to normal dt
                _setload = set([pd.to_datetime(i).date() for i in list(_setload)])

            self.assertEqual(_setload, _setindex)


if __name__ == "__main__":
    unittest.main()
    self = TestIndexer()



"""
Created on Thu Jul 22 16:19:28 2021

@author: DW
"""
from time import sleep
from delegator.main_delegator import MainDelegator


def make_examples():
    _main_run = MainDelegator(run_mode="make_examples")


if __name__ == "__main__":
    print(
        'Hello Docker World, the raman_fitting "make examples" command from the containter is starting in 3 seconds...'
    )
    sleep(2)
    print("\n...and Go!.....\n")
    make_examples()



import unittest
from pathlib import Path

import numpy as np
import pytest

from raman_fitting.datafiles import example_files
from raman_fitting.processing.cleaner import Despiker


class TestDespiker(unittest.TestCase):
    def test_Despiker(self):

        desp = Despiker(np.array([1, 2, 3, 4, 5]))
        self.assertEqual(len(desp.df), 5)

        desp = Despiker(np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
        self.assertEqual(len(desp.df), 10)

        desp = Despiker(np.array([2, 2, 2, 2, 2, 2, 30, 20, 2, 2, 2, 2, 2, 2]))


def _debugging():

    self = TestDespiker()
    desp = Despiker(np.array([2, 2, 2, 2, 2, 2, 30, 20, 2, 2, 2, 2, 2, 2]))
    desp.plot_Z()
    intensity = desp.result["input_intensity"]
    Z_t_filtered = desp.result["Z_t_filtered"]
    moving_window_size = desp.moving_window_size
    Z_threshold = desp.Z_threshold


if __name__ == "__main__":
    unittest.main()



"""
Created on Sun Jun  6 09:35:02 2021

@author: DW
"""

import unittest
from functools import partial

import pytest
from lmfit import Model

import raman_fitting
from raman_fitting.deconvolution_models.base_model import _SUBSTRATE_PEAK, BaseModel

_SUBSTRATE_PREFIX = _SUBSTRATE_PEAK.split("peak")[0]


def _get_list_components(bm):
    _listcompsprefix = partial(map, lambda x,: getattr(x, "prefix"))
    _bm_prefix = list(_listcompsprefix(bm.lmfit_model.components))
    return _bm_prefix


class TestBaseModel(unittest.TestCase):
    def test_empty_base_model(self):
        bm = BaseModel()
        self.assertEqual(bm.model_name, "")
        self.assertFalse(bm.has_substrate)
        bm.add_substrate()
        self.assertIn(bm.model_name, _SUBSTRATE_PEAK)
        self.assertEqual(type(bm.lmfit_model).__qualname__, "GaussianModel")
        self.assertIn(bm.lmfit_model.prefix, _SUBSTRATE_PEAK)
        self.assertTrue(issubclass(type(bm.lmfit_model), Model))

    def test_base_model_2peaks(self):
        bm = BaseModel(model_name="K2+D+G")

        self.assertListEqual(_get_list_components(bm), ["D_", "G_"])
        bm.add_substrate()
        self.assertListEqual(_get_list_components(bm), ["D_", "G_", _SUBSTRATE_PREFIX])
        bm.remove_substrate()
        self.assertListEqual(_get_list_components(bm), ["D_", "G_"])

    def test_base_model_wrong_chars_model_name(self):
        bm = BaseModel(model_name="K2+---////+  +7 +K1111+1D+D2")
        self.assertListEqual(_get_list_components(bm), ["D2_"])
        bm.add_substrate()
        self.assertListEqual(_get_list_components(bm), ["D2_", _SUBSTRATE_PREFIX])


if __name__ == "__main__":
    unittest.main()



# pylint: disable=W0614,W0611,W0622
# flake8: noqa
# isort:skip_file

# Main Loop Delegator
from raman_fitting.delegating.main_delegator import MainDelegator, make_examples

# Indexer
from raman_fitting.indexing.indexer import MakeRamanFilesIndex as make_index


# Processing
from raman_fitting.processing.spectrum_template import SpectrumTemplate
from raman_fitting.processing.spectrum_constructor import (
    SpectrumDataLoader,
    SpectrumDataCollection,
)

# Modelling / fitting
from raman_fitting.deconvolution_models.fit_models import InitializeModels, Fitter

# Exporting / Plotting
from raman_fitting.exporting.exporter import Exporter
from raman_fitting.config import filepath_settings
